Package: utsf
Title: Univariate Time Series Forecasting
Version: 1.3.4
Authors@R: c(
    person("Maria Pilar", "Frias-Bustamante", , "mpfrias@ujaen.es", role = "aut",
           comment = c(ORCID = "0000-0001-6886-0953")),
    person("Francisco", "Martinez", , "fmartin@ujaen.es", role = c("aut", "cre", "cph"),
           comment = c(ORCID = "0000-0002-5206-1898"))
  )
Description: An engine for univariate time series forecasting using
    different regression models in an autoregressive way. The engine
    provides an uniform interface for applying the different models. 
    Furthermore, it is extensible so that users can easily apply their
    own regression models to univariate time series forecasting and 
    benefit from all the features of the engine, such as preprocessings
    or estimation of forecast accuracy.
Maintainer: Francisco Martinez <fmartin@ujaen.es>
License: MIT + file LICENSE
URL: https://github.com/franciscomartinezdelrio/utsf
BugReports: https://github.com/franciscomartinezdelrio/utsf/issues
Imports: Cubist, FNN, forecast, generics, ggplot2, ipred, methods,
        ranger, rpart, vctsfr, xgboost
Suggests: knitr, nnet, randomForest, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
RoxygenNote: 8.0.0
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-07-21 07:45:22 UTC; UJA
Author: Maria Pilar Frias-Bustamante [aut] (ORCID:
    <https://orcid.org/0000-0001-6886-0953>),
  Francisco Martinez [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-5206-1898>)
Repository: CRAN
Date/Publication: 2026-07-21 08:30:02 UTC
Built: R 4.6.1; ; 2026-07-21 12:25:47 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(autoplot,utsf_forecast)
S3method(forecast,utsf)
S3method(plot,utsf_forecast)
S3method(predict,utsf)
S3method(print,fd)
S3method(print,summary.utsf)
S3method(print,utsf)
S3method(print,utsf_forecast)
S3method(summary,utsf)
export(build_examples)
export(create_model)
export(efa)
export(forecast)
export(tune_grid)
importFrom(generics,accuracy)
importFrom(generics,forecast)
importFrom(ggplot2,autoplot)
importFrom(graphics,plot)

# utsf

<!-- badges: start -->

[![R-CMD-check](https://github.com/franciscomartinezdelrio/utsf/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/franciscomartinezdelrio/utsf/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The utsf package provides an engine for univariate time series
forecasting using different regression models in an autoregressive way.
The engine provides an uniform interface for applying the different
models. Furthermore, it is extensible so that users can easily apply
their own regression models to univariate time series forecasting and
benefit from all the features of the engine, such as preprocessings or
estimation of forecast accuracy.

## Installation

You can install the **development** version of utsf from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("franciscomartinezdelrio/utsf")
```

or you can install the **stable** version from
[CRAN](https://cran.r-project.org/):

``` r
install.packages("utsf")
```

## Example

This is a basic example which shows you how to solve a common problem:

``` r
library(utsf)
# Forecast the next 12 future values of time series UKDriverDeaths using random forest
m <- create_model(UKDriverDeaths, method = "rf")
f <- forecast(m, h = 12)
f$pred # to see the forecast
#>           Jan      Feb      Mar      Apr      May      Jun      Jul      Aug
#> 1985 1299.725 1215.264 1227.347 1170.516 1261.783 1263.307 1325.378 1346.959
#>           Sep      Oct      Nov      Dec
#> 1985 1412.481 1543.708 1698.649 1805.963
library(ggplot2)
autoplot(f)
```

<img src="man/figures/README-example-1.png" alt="" width="100%" />

If you are interested in this package you can read its
[vignette](https://cran.r-project.org/package=utsf) where all its
important features are described.

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(utsf)

## -----------------------------------------------------------------------------
m <- create_model(AirPassengers, lags = 1:12, method = "rt")
f <- forecast(m, h = 12)

## -----------------------------------------------------------------------------
f$pred
library(ggplot2)
autoplot(f)

## -----------------------------------------------------------------------------
head(m$targets)  # first targets
head(m$features) # and its associated features

## -----------------------------------------------------------------------------
t <- ts(c(1, 3, 6, 7, 9, 11, 16))
out <- create_model(t, lags = c(1, 2, 4), trend = "none")
cbind(out$features, Target = out$targets)

## -----------------------------------------------------------------------------
m <- create_model(USAccDeaths, method = "mt")
f <- forecast(m, h = 12, PI = TRUE, level = 90)
f
library(ggplot2)
autoplot(f)

## -----------------------------------------------------------------------------
m <- create_model(fdeaths, lags = 1:12, method = "rt")
m$model

## -----------------------------------------------------------------------------
# Modeling function: in this case (k-NN) just stores the training set
my_knn_model <- function(X, y, ...) { structure(list(X = X, y = y), class = "my_knn") }

# Regression function
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
}

m <- create_model(AirPassengers, lags = 1:12, method = my_knn_model)
f <- forecast(m, h = 12)
f$pred
autoplot(f)

## -----------------------------------------------------------------------------
# Modeling function
my_knn_model2 <- function(X, y, ...) {
  structure(list(X = X, y = y), class = "my_knn2")
}

# Regression function
predict.my_knn2 <- function(object, new_value) {
  k <- 3 # number of nearest neighbors
  distances <- sapply(1:nrow(object$X), function(i) sum((object$X[i, ] - new_value)^2))
  k_nearest <- order(distances)[1:k]
  mean(object$y[k_nearest])
}

m2 <- create_model(AirPassengers, lags = 1:12, method = my_knn_model2)
forecast(m2, h = 12)$pred

## -----------------------------------------------------------------------------
library(randomForest)
my_model <- function(X, y, ...) { randomForest(x = X, y = y) }
m <- create_model(USAccDeaths, lags = 1:12, method = my_model)
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)
print(m$model)

## -----------------------------------------------------------------------------
library(nnet)
my_model <- function(X, y, ...) {
  nnet(x = X, y = y, size = 5, linout = TRUE, trace = FALSE)
}
m <- create_model(USAccDeaths, lags = 1:12, method = my_model)
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)

## -----------------------------------------------------------------------------
# A bagging model set with default parameters
m <- create_model(USAccDeaths, lags = 1:12, method = "bagging")
length(m$model$mtrees) # number of regression trees (25 by default)
# A bagging model set with 3 regression trees
m2 <- create_model(USAccDeaths,
                   lags = 1:12,
                   method = "bagging",
                   nbagg = 3
)
length(m2$model$mtrees) # number of regression trees

## -----------------------------------------------------------------------------
# Function to train the model
my_knn_model <- function(X, y, ...) {
  param <- list(...)
  k <- if ("k" %in% names(param)) param$k else 3
  structure(list(X = X, y = y, k = k), class = "my_knn")
}

# Regression function for object of class my_knn
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value,
               y = object$y, k = object$k)$pred
}

# The model is trained with default parameters (k = 3)
m <- create_model(AirPassengers, lags = 1:12,  method = my_knn_model)
print(m$model$k)
# The model is trained with k = 5
m2 <- create_model(AirPassengers, method = my_knn_model, k = 5)
print(m2$model$k)

## -----------------------------------------------------------------------------
library(randomForest)
my_model <- function(X, y, ...) {
    param <- list(...)
    args <- list(x = X, y = y, ntree = 200) # default parameters
    args <- args[!(names(args) %in% names(param))]
    args <- c(args, param)
    do.call(randomForest::randomForest, args = args)
}
# The random forest is built with our default parameters
m <- create_model(USAccDeaths, method = my_model)
print(m$model$ntree)
print(m$model$mtry)
# The random forest is built with ntree = 400 and mtry = 6
m <- create_model(USAccDeaths, method = my_model, ntree = 400, mtry = 6)
print(m$model$ntree)
print(m$model$mtry)

## -----------------------------------------------------------------------------
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- efa(m, h = 4, type = "normal", size = 8)
r$per_horizon
r$global

## ----out.width = '85%', echo = FALSE------------------------------------------
knitr::include_graphics("ro_normal.png")

## ----eval = FALSE-------------------------------------------------------------
# m <- create_model(UKgas, lags = 1:4, method = "knn")
# # Use the last 9 observations of the series to build test sets
# r1 <- efa(m, h = 4, type = "normal", size = 9)
# # Use the last 20% observations of the series to build test sets
# r2 <- efa(m, h = 4, type = "normal", prop = 0.2)

## -----------------------------------------------------------------------------
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- efa(m, h = 4, type = "minimum")
r$per_horizon
r$global

## ----out.width = '85%', echo = FALSE------------------------------------------
knitr::include_graphics("ro_minimum.png")

## -----------------------------------------------------------------------------
timeS <- ts(1:25)
m <- create_model(timeS, lags = 1:3, method = "mt")
r <- efa(m, h = 5, size = 7)
r$test_sets
r$predictions

## -----------------------------------------------------------------------------
timeS <- ts(1:25)
m <- create_model(timeS, lags = 1:3, method = "mt")
r <- efa(m, h = 3, type = "minimum")
r$test_sets
r$predictions

## -----------------------------------------------------------------------------
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- tune_grid(m, h = 4, tuneGrid = expand.grid(k = 1:7), type = "normal", size = 8)
# To see the estimated forecast accuracy with the different configurations
r$tuneGrid

## -----------------------------------------------------------------------------
r$best
r$forecast

## -----------------------------------------------------------------------------
plot(r$tuneGrid$k, r$tuneGrid$RMSE, type = "o", pch = 19, xlab = "k (number of nearest neighbors)", ylab = "RMSE", main = "Estimated accuracy")

## -----------------------------------------------------------------------------
m <- create_model(UKDriverDeaths, lags = 1:12, method = "rf")
r <- tune_grid(m, 
               h = 12, 
               tuneGrid = expand.grid(num.trees = c(200, 500), replace = c(TRUE, FALSE), mtry = c(4, 8)),
               type = "normal", 
               size = 12
)
r$tuneGrid

## -----------------------------------------------------------------------------
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "none")
f <- forecast(m, h = 4)
autoplot(f)

## -----------------------------------------------------------------------------
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "differences", nfd = 1)
f <- forecast(m, h = 4)
autoplot(f)

## -----------------------------------------------------------------------------
# The order of first differences is estimated using the ndiffs function
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "differences", nfd = -1)
m$differences

## -----------------------------------------------------------------------------
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "none")
cbind(m$features, Targets = m$targets)

## -----------------------------------------------------------------------------
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "additive", transform_features = FALSE)
cbind(m$features, Targets = m$targets)

## -----------------------------------------------------------------------------
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "additive", transform_features = TRUE)
cbind(m$features, Targets = m$targets)

## -----------------------------------------------------------------------------
m <- create_model(airmiles, lags = 1:4, method = "rf")
f <- forecast(m, h = 4)
autoplot(f)

## -----------------------------------------------------------------------------
t <- ts(10 * 1.05^(1:20))
m_m <- create_model(t, lags = 1:3, method = "rf", trend = "multiplicative")
f_m <- forecast(m_m, h = 4)
m_a <- create_model(t, lags = 1:3, method = "rf", trend = "additive")
f_a <- forecast(m_a, h = 4)
library(vctsfr)
plot_predictions(t, predictions = list(Multiplicative = f_m$pred, Additive = f_a$pred))

## ----fig.width= 6-------------------------------------------------------------
m_m <- create_model(AirPassengers, lags = 1:12, method = "knn", trend = "multiplicative")
f_m <- forecast(m_m, h = 36)
m_a <- create_model(AirPassengers, lags = 1:12, method = "knn", trend = "additive")
f_a <- forecast(m_a, h = 36)
library(vctsfr)
plot_predictions(AirPassengers, predictions = list(Multiplicative = f_m$pred, Additive = f_a$pred))

build_model <- function(X, y, method, param) {
  if (method == "knn") {
    k <- if ("k" %in% names(param)) param[["k"]] else 3
    if (length(y) < k) {
      m <- paste("It is not possible to use the method KNN with k =", k,
                 "and a training set of size", length(y))
      stop(m)
    }
    model <- paste0("K-nearest neighbors with k = ", k)
  } else if (method == "lm") {
    args <- list(formula = y ~ ., data = data.frame(cbind(X, y = y)))
    args <- c(args, param)
    model <- do.call(stats::lm, args = args)
  } else if (method == "rt") {
    df <- cbind(X, targets = y)
    args <- list(formula = targets ~ .,
                 data = df,
                 method = "anova")
    args <- c(args, param)
    model <- do.call(rpart::rpart, args = args)
  } else if (method == "mt") {
    args <- list(x = as.data.frame(X), y = y, committees = 5)
    args <- args[!(names(args) %in% names(param))]
    args <- c(args, param)
    model <- do.call(Cubist::cubist, args = args)
  }  else if (method == "bagging") {
    df <- cbind(X, targets = y)
    args <- list(formula = targets ~ ., data = df)
    args <- c(args, param)
    model <- do.call(ipred::bagging, args = args)
  } else if (method == "rf") { # random forest
    df <- cbind(X, targets = y)
    args <- list(formula = targets ~ .,
                 data = df,
                 oob.error = FALSE,
                 mtry = floor((ncol(df)-1)/3)
    )
    args <- args[!(names(args) %in% names(param))]
    args <- c(args, param)
    model <- do.call(ranger::ranger, args = args)
  } else if (method == "xgboost") { # extreme gradient boosting
    args <- list(x = as.matrix(X), y = y)
    args <- c(args, param)
    model <- do.call(xgboost::xgboost, args = args)
  }
  model
}

#' Prediction from `utsf` objects
#'
#' Predict the class of a new observation based on the model associated with the
#' `utsf` object
#' @param object object of class `utsf`.
#' @param new_value a data frame with one row of a new observation.
#' @param ... further arguments passed to or from other methods.
#'
#' @return a numeric value with the forecast.
#' @export
predict.utsf <- function(object, new_value, ...) {
  if (object$method == "knn") {
    check_param(object, FNN::knn.reg, "FNN::knn.reg")
    args <- list(train = object$features,
                 test = new_value,
                 y = object$targets)
    args <- c(args, object$param)
    return(do.call(FNN::knn.reg, args = args)$pred)
  } else if (object$method == "rf") {
    return(stats::predict(object$model, new_value)$predictions)
  } else {
    # k <- 3
    # query <- as.matrix(new_value)
    # m <- as.matrix(object$features)
    # colnames(m) <- NULL
    # indices <- FNN::get.knnx(m, query = query, k = k)$nn.index[1, ]
    # k_nearest <- rbind(object$features[indices, ], new_value)
    # return(mean(stats::predict(object$model, k_nearest)))
    return(stats::predict(object$model, new_value))
  }
}

# Check that parameters provided by the user for customizing model building belong 
# to building function. If not, execution is stopped
# param object An object of class utsf
# param f An object of class function. The function 
# param fname A string. The name of the function
check_param <- function(object, f, fname) {
  formal <- methods::formalArgs(f)
  dif <- setdiff (names(object$param), formal)
  if (length(dif) > 0)
    stop(paste("Error in \"param\" argument, parameters", dif, "are not part of",
               fname, "function"))
}

# simulate a path
sim <- function(model, h, errors) {
  ts <- model$ts
  prediction <- numeric(h)
  values <- as.vector(ts)
  standard_d <- stats::sd(errors)
  for (hor in 1:h) {
    example <- values[(length(values) + 1) - rev(model$lags)]
    prediction[hor] <- predict_one_value_transforming(model, example) + 
      stats::rnorm(1, mean = 0, sd = standard_d)
    values <- c(values, prediction[hor])
  }
  temp <- stats::ts(1:2,
                    start = stats::end(ts),
                    frequency = stats::frequency(ts)
  )
  prediction <- stats::ts(prediction,
                          start = stats::end(temp),
                          frequency = stats::frequency(ts)
  )
  if (model$trend == "differences" && model$differences$differences > 0) {
    prediction <- fd_unpreprocessing(prediction, model$differences)
  }
  prediction
}

simulations <- function(model, h, level) {
  r <- efa(model, h, size = h + 10)
  errors <- r$test_sets - r$predictions
  m <- replicate(1000, sim(model, h, errors))
  d <- (100-level) / 2
  apply(m, 1, function(x) stats::quantile(x, probs = c(d/100, (100-d)/100)))
}

#'Estimate the forecast accuracy of a model on a time series
#'
#'It uses an object of class `utsf` to asses the forecasting accuracy of its
#'associated model on its associated time series applying a rolling origin
#'evaluation.
#'
#'@param model An object of class `utsf` with a model trained with a time
#'  series.
#'@param h A positive integer. The forecasting horizon.
#'@param type A string. Possible values are `"normal"` (the default) and
#'  `"minimum"`. See the vignette [utsf][utsf] for an explanation of both ways
#'  of evaluating forecast accuracy.
#'@param size An integer. It is the size of the test set (how many of the last
#'  observations of the time series are used as test set). It can only be used
#'  when the type parameter is `"normal"`. By default, it is the length of the
#'  forecasting horizon.
#'@param prop A numeric value in the range (0, 1). It is the proportion of the
#'  time series used as test set. It can only be used when the type parameter is
#'  `"normal"`.
#'
#'@returns A list with four components: \item{`per_horizon`}{A matrix with the
#'  estimated forecast accuracy per forecasting horizon using several
#'  forecasting accuracy measures.}
#' \item{`global`}{The average estimated forecast accuracy for all the horizons. It is computed as the mean
#' of the different rows of the `per_horizon` component.}
#'  \item{`test_sets`}{A matrix with the test sets used in the evaluation. Each
#'  row of the matrix is a test set.} \item{`predictions`}{The predictions for the test sets.}
#'@export
#'
#' @examples
#' m <- create_model(UKgas, lags = 1:4, method = "rt")
#' efa(m, h = 4, type = "normal", size = 8)
efa <- function(model, h, type = c("normal", "minimum"), size = NULL, prop = NULL) {
  # Check model parameter
  if (!inherits(model, "utsf"))
    stop("model parameter should be of class 'utsf'")
  # Check h parameter
  if (! (is.numeric(h) && length(h) == 1 && h >= 1 && floor(h) == h))
    stop("h parameter should be an integer scalar greater than zero")
  # Check type parameter
  type <- match.arg(type)
  if (type == "normal" && !is.null(size) && !is.null(prop))
    stop("Only one of size and prop parameters should be different from NULL")
  # Check size parameter
  if (!is.null(size) && (length(size) > 1 || !is.numeric(size) || round(size) != size || size <= 0))
    stop("size parameter should be a positive integer scalar")
  # Check prop parameter
  if (!is.null(prop) && (length(prop) > 1 || !is.numeric(prop) || prop <= 0 || prop >= 1))
    stop("prop parameter should be a number in the range (0, 1)")
  
  if (length(model$ts) <= h)
    stop("Time series is too short to estimate forecast accuracy")

  if (type == "normal") {
    if (is.null(size) && is.null(prop)) {
      size <- h
    } else {
      size <- if (!is.null(size)) size else trunc(prop*length(model$ts))
    }
    if (!is.null(model$lags) && max(model$lags) >= length(model$ts)-size)
      stop("Time series is too short to estimate forecast accuracy")
    
    if (is.character(model$method) && model$method == "knn" && !is.null(model$lags)) {
      k <- if ("k" %in% names(model$param)) model$param[["k"]] else 3
      if (length(model$ts)-size-max(model$lags) < k)
        stop("Time series is too short to estimate forecast accuracy")
    }
    test_sets <- matrix(NA, nrow = size - h + 1, ncol = h)
    colnames(test_sets) <- paste0("h=", 1:h)
    predictions <- test_sets
    row <- 1
    for (position in (length(model$ts)-size+1):(length(model$ts)-h+1)) {
      tt <- training_test2(model$ts, position, h)
      test_sets[row, 1:h] <- tt$test
      l <- as.list(model$call)[-1]
      l$timeS <- tt$training
      m <- do.call("create_model", args = l)
      predictions[row, 1:h] <- forecast(m, h = h)$pred
      row <- row + 1
    }
  } else { # type is minimum
    if (!is.null(model$lags) && max(model$lags) >= length(model$ts)-h)
      stop("Time series is too short to estimate forecast accuracy")
    if (is.character(model$method) && model$method == "knn" && !is.null(model$lags)) {
      k <- if ("k" %in% names(model$param)) model$param[["k"]] else 3
      if (length(model$ts)-h-max(model$lags) < k)
        stop("Time series is too short to estimate forecast accuracy")
    }
    test_sets <- matrix(NA, nrow = h, ncol = h)
    predictions <- test_sets
    for (hor in 1:h) {
      tt <- training_test(model$ts, hor)
      test_sets[hor, 1:hor] <- tt$test
      l <- as.list(model$call)[-1]
      l$timeS <- tt$training
      m <- do.call("create_model", args = l)
      predictions[hor, 1:hor] <- forecast(m, h = hor)$pred
    }
  }
  errors <- test_sets - predictions
  global <- rep(0, 4)
  efa_per_horizon <- matrix(0, nrow = 4, ncol = h)
  rownames(efa_per_horizon) <- c("MAE", "MAPE", "sMAPE", "RMSE")
  colnames(efa_per_horizon) <- paste("Horizon", 1:h)
  efa_per_horizon["MAE", ] <- colMeans(abs(errors), na.rm = TRUE)
  efa_per_horizon["MAPE", ] <- colMeans(100*abs((test_sets - predictions) / test_sets), na.rm = TRUE)
  efa_per_horizon["sMAPE", ] <- colMeans(abs(test_sets - predictions) / (abs(test_sets)+abs(predictions))*200,
                                         na.rm = TRUE)
  efa_per_horizon["RMSE", ] <- sqrt(colMeans(errors^2, na.rm = TRUE))
  names(global) <- c("MAE", "MAPE", "sMAPE", "RMSE")
  list(per_horizon = efa_per_horizon, 
       global = rowMeans(efa_per_horizon),
       test_sets = test_sets,
       predictions = predictions
  )
}

training_test <- function(timeS, n) {
  training <- stats::ts(utils::head(timeS, -n),
                        start = stats::start(timeS),
                        frequency = stats::frequency(timeS)
  )
  test <- stats::ts(utils::tail(timeS, n),
                    end = stats::end(timeS),
                    frequency = stats::frequency(timeS)
  )
  list(training = training, test = test)
}

# position is the index where the test size begins
training_test2 <- function(timeS, position, test_set_size) {
  training <- stats::ts(utils::head(timeS, position-1),
                        start = stats::start(timeS),
                        frequency = stats::frequency(timeS)
  )
  test <- stats::ts(timeS[position:(position + test_set_size - 1)],
                    start = stats::time(timeS)[position],
                    frequency = stats::frequency(timeS)
  )
  list(training = training, test = test)
}

#' @export
print.utsf <- function (x, ...) {
  cat("\nCall:  ",
      paste(deparse(x$call),
            sep = "\n",
            collapse = "\n"
      ),
      "\n\n",
      sep = ""
  )
  cat("Autoregressive lags:", x$lags, "\n")
  if (x$trend == "additive") {
    cat ("Additive tranformation applied to", 
         if (x$transform_features) "features and", "targets\n")
  }
  if (x$trend == "multiplicative") {
    cat ("Multiplicative tranformation applied to", 
         if (x$transform_features) "features and", "targets\n")
  }
  if (x$trend == "differences") {
    cat("First differences applied as preprocessing.", nd2character(x$differences), "\n")
  }
  cat("Regression model: ")
  if (inherits(x$method, "function")) {
    cat("provided by user\n")
  } else {
    method <-  switch(x$method,
                      "knn" = "k-nearest neighbors",
                      "lm" = "linear model",
                      "rt" = "regression tree",
                      "mt" = "model tree",
                      "rf" = "random forest",
                      "bagging" = "bagging"
    )
    cat(method, "\n")
  }
    
  # if (is.null(x$global_efa) && is.null(x$tuneGrid)) {
  #   cat("Forecast:\n")
  #   print(x$pred)
  # } else if (!is.null(x$global_efa)) {
  #   cat("Estimated average forecast accuracy for horizon ", length(x$pred), ":\n", sep = "")
  #   print(x$global_efa)
  # }  else if (!is.null(x$tuneGrid)) {
  #   cat("Estimated average forecast accuracy for different combinations of tuning parameters:\n")
  #   print(x$tuneGrid)
  #   minimum <- which.min(x$tuneGrid$RMSE)
  #   cat("\nBest combination according to RMSE:\n")
  #   print(x$tuneGrid[minimum, ])
  # }
  invisible(x)
}

#' @export
summary.utsf <- function (object, ...) structure(object, class = "summary.utsf")

#' @export
print.summary.utsf <- function (x, ...) {
  cat("\nCall:  ",
      paste(deparse(x$call),
            sep = "\n",
            collapse = "\n"
      ),
      "\n\n",
      sep = ""
  )
  cat("Autoregressive lags:", x$lags, "\n")
  if (x$trend == "additive") {
    cat ("Additive tranformation applied to", 
         if (x$transform_features) "features and", "targets\n")
  }
  if (x$trend == "multiplicative") {
    cat ("Multiplicative tranformation applied to", 
         if (x$transform_features) "features and", "targets\n")
  }
  if (x$trend == "differences") {
    cat("First differences applied as preprocessing.", nd2character(x$fd), "\n")
  }
  cat("Regression model: ")
  if (inherits(x$method, "function")) {
    cat("provided by user\n")
  } else {
    method <-  switch(x$method,
                      "knn" = "k-nearest neighbors",
                      "lm" = "linear model",
                      "rt" = "regression trees",
                      "mt" = "model trees",
                      "rf" = "random forest",
                      "bagging" = "bagging"
    )
    cat(method, "\n")
  }
  invisible(x)
}

#' @export
print.utsf_forecast <- function (x, ...) {
  if (is.null(x$lower)) {
    print(x$pred)
  } else {
    l = paste("Lo", x$level)
    h <- paste("Hi", x$level)
    o <- cbind(x$pred, l = x$lower, h = x$upper)
    colnames(o) <- c("Point Forecast", l, h)
    print(o)
  }
  invisible(x)
}

#'Estimate the forecast accuracy of a model on a time series according to a grid
#'of parameters
#'
#'It uses an object of class `utsf` to asses the forecasting accuracy of its
#'associated model on its associated time series applying rolling origin
#'evaluation according to different configurations of model parameters.
#'
#'The estimation of forecast accuracy is done with the [efa()] function. The
#'best combination of parameters is used to train the model with all the
#'historical values of the time series and forecast `h` values ahead.
#'@inheritParams efa
#'@param tuneGrid A data frame with possible tuning values. The columns are
#'  named as the tuning parameters.
#'@returns A list with three components: \item{`tuneGrid`}{A data frame with the
#'  different combination of parameters and the estimated forecast accuracy of a
#'  model trained with those parameters.} \item{`best`}{The best combination of
#'  parameters according to root mean squared error.} \item{`forecast`}{An
#'  object of class `utsf_forecast` with the forecast for horizon `h` using the
#'  best estimated combination of parameters.}
#'@export
#'
#' @examples
#' m <- create_model(UKgas, lags = 1:4, method = "knn")
#' tune_grid(m, h = 4, tuneGrid = expand.grid(k = 1:7), type = "normal", size = 8)
tune_grid <- function(model, h, tuneGrid, type = c("normal", "minimum"), size = NULL, prop = NULL) {
  # Check model parameter
  if (!inherits(model, "utsf"))
    stop("model parameter should be of class 'utsf'")

    # Check h parameter
  if (! (is.numeric(h) && length(h) == 1 && h >= 1 && floor(h) == h))
    stop("h parameter should be an integer scalar greater than zero")

  # Check tuneGrid parameter
  if (! is.data.frame(tuneGrid))
    stop("parameter tuneGrid should be a data frame")
  
  # Check type parameter
  type <- match.arg(type)
  if (type == "normal" && !is.null(size) && !is.null(prop))
    stop("Only one of size and prop parameters should be different from NULL")
  # Check size parameter
  if (!is.null(size) && (length(size) > 1 || !is.numeric(size) || round(size) != size || size <= 0))
    stop("size parameter should be a positive integer scalar")
  # Check prop parameter
  if (!is.null(prop) && (length(prop) > 1 || !is.numeric(prop) || prop <= 0 || prop >= 1))
    stop("prop parameter should be a number in the range (0, 1)")
  
  if (length(model$ts) <= h)
    stop("Time series is too short to estimate forecast accuracy")

  output <- NULL
  for (r in 1:nrow(tuneGrid)) {
    l <- as.list(model$call)[-1]
    l <- c(l, as.list(tuneGrid[r, , drop = FALSE]))
    m <- do.call("create_model", args = l)
    r <- efa(m, h = h, type = type, size = size, prop = prop)
    output <- rbind(output, r$global)
  }
  output <- cbind(tuneGrid, output)
  rownames(output) <- NULL
  best <- as.list(output[which.min(output$RMSE), 1:ncol(tuneGrid), drop = FALSE])
  l <- c(as.list(model$call)[-1], best)
  m <- do.call("create_model", args = l)
  f <- forecast(m, h = h)
  list(tuneGrid = output, best = best, forecast = f)
}
  

# Differences preprocessing
# timeS is the time series
# differences is the number of differences, if -1, the number of differences is
# estimated with the function forecast::ndiffs
fd_preprocessing <- function(timeS, differences = -1) {
  prepro <- list()
  prepro$asked_differences <- differences
  if (differences == -1) {
    differences <- forecast::ndiffs(timeS)
  }
  if (differences == 0) {
    prepro$differences <- 0
    return(structure(prepro, class = "fd"))
  }
  prepro$original <- timeS
  
  last_values <- utils::tail(timeS, differences)
  timeS <- diff(timeS, differences = differences)
  prepro$last_values <- last_values
  prepro$differences <- differences
  
  prepro$preprocessed <- timeS
  structure(prepro, class = "fd")
}

# first differences unpreprocessing
# forecast: a forecast based on the first differenced series
# prepro: information about the first differences preprocessing
fd_unpreprocessing <- function(forecast, prepro) {
  temp <- stats::diffinv(forecast, differences = prepro$differences, xi = prepro$last_values)
  utils::tail(temp, -prepro$differences)
}

nd2character <- function(p) {
  o <- "Order of first differences"
  if (p$asked_differences == -1) o <- paste(o, "(selected automatically)")
  paste0(o, ": ", p$differences)
}
#' @export
print.fd <- function(x, ...) {
  cat(nd2character(x), "\n")
  invisible(x)
}

# @param object S3 object of class utsf
# @param h forecasting horizon (a positive integer)
recursive_prediction <- function(object, h) {
  ts <- object$ts
  prediction <- numeric(h)
  values <- as.vector(ts)
  # tiempos <- stats::time(stats::ts(1:(length(ts)+h), start = start(ts), frequency = frequency(ts))) #añadido
  # ciclo <- stats::cycle(stats::ts(1:(length(ts)+h), start = start(ts), frequency = frequency(ts))) # Quitar
  for (hor in 1:h) {
    example <- values[(length(values) + 1) - rev(object$lags)]
    #example[length(example) + 1] <- tiempos[length(values) + 1 - rev(object$lags)[1]] # añadido
    # example[length(example) + 1] <- ciclo[length(ts)+hor] # Quitar
    prediction[hor] <- predict_one_value_transforming(object, example)
    values <- c(values, prediction[hor])
  }
  temp <- stats::ts(1:2,
                    start = stats::end(ts),
                    frequency = stats::frequency(ts)
  )
  prediction <- stats::ts(prediction,
                          start = stats::end(temp),
                          frequency = stats::frequency(ts)
  )
  prediction
}

#'Train an univariate time series forecasting model
#'
#'This function trains a model from the historical values of a time series using
#'an autoregressive approach: the targets are the historical values and the
#'features of the targets their lagged values.
#'
#'The functions used to build and train the model are:
#' * KNN: In this case no model is built and the function [FNN::knn.reg()] is
#'used to predict the future values of the time series. By default, k is equal
#'to 3.
#' * Linear models: Function [stats::lm()] to build the model and the method
#'[stats::predict.lm()] associated with the trained model to forecast the future
#'values of the time series.
#' * Regression trees: Function [rpart::rpart()] to build the model and the
#'method [rpart::predict.rpart()] associated with the trained model to forecast
#'the future values of the time series.
#' * Model trees: Function [Cubist::cubist()] to build the model and the
#'method [Cubist::predict.cubist()] associated with the trained model to
#'forecast the future values of the time series. By default, the parameter
#'`committees` is set to 5.
#' * Bagging: Function [ipred::bagging()] to build the model and the
#'method [ipred::predict.regbagg()] associated with the trained model to
#'forecast the future values of the time series.
#' * Random forest: Function [ranger::ranger()] to build the model and the
#'method [ranger::predict.ranger()] associated with the trained model to
#'forecast the future values of the time series.
#' * Extreme gradient boosting: Function [xgboost::xgboost()] to build the model and the
#'method [xgboost::predict.xgboost()] associated with the trained model to
#'forecast the future values of the time series.
#'
#'@param timeS A time series of class `ts` or a numeric vector.
#'@param lags An integer vector, in increasing order, expressing the lags used
#'  as autoregressive variables. If the default value (`NULL`) is provided, a
#'  suitable vector is chosen.
#'@param method A string indicating the method used for training and
#'  forecasting. Allowed values are:
#'   * `"knn"`: k-nearest neighbors (the default)
#'   * `"lm"`: linear regression
#'   * `"rt"`: regression trees
#'   * `"mt"`:  model trees
#'   * `"bagging"`
#'   * `"rf"`: random forests
#'   * `"xgboost"`: extreme gradient boosting
#'
#'  See details for a brief explanation of the models. It is also possible to
#'  use your own regression model, in that case a function explaining how to
#'  build your model must be provided, see the vignette for further details.
#'
#'@param trend A character indicating the type of preprocessing applied to the
#'  time series in order to deal with trending series, see the vignette for
#'  details.
#'
#'@param nfd In the case that the parameter `trend` has the value "differences",
#'  it specifies the order of first differences to be applied. If the default
#'  (-1) is used, the order of first differences needed by the time series will
#'  be estimated by the [forecast::ndiffs()] function.
#'
#'@param transform_features A logical value indicating whether the training
#'  features are also transformed if the additive or multiplicative
#'  transformation has been used as preprocessing to deal with trending series.
#'
#'@param ... Parameters for the underlying function that builds the model. If no
#'  parameters are provided, the model is normally fitted with its default
#'  parameters. See details for the functions used to train the models.
#'
#'@returns An S3 object of class `utsf`, basically a list with, at least, the
#'  following components: \item{`ts`}{The time series being forecast.}
#'  \item{`features`}{A data frame with the features of the training set. The
#'   column names of the data frame indicate the autoregressive lags.}
#'  \item{`targets`}{A vector with the targets of the training set.}
#'  \item{`lags`}{An integer vector with the autoregressive lags.}
#'  \item{`model`}{The regression model used recursively to make the forecast.}
#'@export
#'
#' @examples
#' ## Build model using k-nearest neighbors
#' create_model(AirPassengers, method = "knn")
#'
#' ## Using k-nearest neighbors changing the default k value
#' create_model(AirPassengers, method = "knn", k = 5)
#'
#' ## Using your own regression model
#'
#' # Function to build the regression model
#' my_knn_model <- function(X, y, param) {
#'   structure(list(X = X, y = y), class = "my_knn")
#'}
#' # Function to predict a new example
#' predict.my_knn <- function(object, new_value) {
#'   FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
#' }
#' create_model(AirPassengers, method = my_knn_model)
#'
create_model <- function(timeS, 
                         lags = NULL, 
                         method = c("knn", "lm", "rt", "mt", "bagging", "rf", "xgboost"), 
                         trend = c("additive", "multiplicative", "differences", "none"),
                         nfd = -1,
                         transform_features = TRUE,
                         ...) {
  # Check timeS parameter
  if (! (stats::is.ts(timeS) || is.vector(timeS, mode = "numeric")))
    stop("timeS parameter should be of class ts or a numeric vector")
  if (! stats::is.ts(timeS))
    timeS <- stats::as.ts(timeS)
  
  trend <- match.arg(trend)
  
  # Check lags parameter
  lagsc <- lags
  if (! (is.null(lagsc) || is.vector(lagsc, mode = "numeric"))) {
    stop("lags parameter should be NULL or numeric")
  }
  if (!is.null(lagsc) && !all(lagsc == floor(lagsc))) {
    stop("lag values should be integer")
  }
  if (is.null(lagsc)) {
    if (stats::frequency(timeS) > 1) {
      lagsc <- 1:stats::frequency(timeS)
    } else {
      partial <- stats::pacf(timeS, plot = FALSE)
      lagsc <- which(partial$acf > 2/ sqrt(length(timeS)))
      if (length(lagsc) == 0 ||
          (length(lagsc) == 1 &&
           trend %in% c("additive", "multiplicative"))) {
        lagsc <- 1:5
      }
    }
  }
  if (is.unsorted(lagsc)) lagsc <- sort(lagsc)
  if (lagsc[1] < 1) stop("lag values should be greater than zero")
  
  if ((length(lagsc) == 1 && trend %in% c("additive", "multiplicative")) &&  transform_features) {
    stop("It does not make sense to use only 1 autoregressive lag with the additive or multiplicative transformation of features")
  }
  
  if (utils::tail(lagsc, 1) >= length(timeS)) {
    stop("Maximum lag cannot be greater or equal to the length of the series")
  }
  
  # Check method parameter
  tryCatch(method <- match.arg(method),
           error = function(cond) {
             if (length(method) != 1)
               stop("parameter method cannot be a vector")
             if (! class(method) %in% "function")
               stop("parameter method should be a function or valid method")
             method <- method
           }
  )

  # Check nfd parameter
  if (trend == "differences") {
    if (! (is.numeric(nfd) && length(nfd) == 1 && nfd >= -1 && floor(nfd) == nfd))
      stop("nfd parameter should be an integer scalar value >= -1")
  }
  
  # Check transform_features parameter
  if (! (is.logical(transform_features) && length(transform_features) == 1))
    stop("transform_features parameter should be a logical value")
  
  
  # Create training set and targets / transformations / preprocessing
  if (trend == "differences") {
    preprocessing_fd <- fd_preprocessing(timeS, nfd)
    if (preprocessing_fd$differences == 0) {
      out <- build_examples(timeS, rev(lagsc))
    } else {
      out <- build_examples(preprocessing_fd$preprocessed, rev(lagsc))
    }
    out$differences <- preprocessing_fd
  } else {
    out <- build_examples(timeS, rev(lagsc))
    if (trend == "additive") {
      means <- rowMeans(out$features)
      if (transform_features) { 
        out$features <- sapply(1:nrow(out$features), 
                               function(row) out$features[row, ] - means[row])
        out$features <- t(out$features)
      }
      out$features <- as.data.frame(out$features)
      out$targets <- out$targets - means
      # means <- rowMeans(out$features[, 1:length(lagsc)])
      # out$features[ , 1:length(lagsc)] <- sapply(1:nrow(out$features),
      #                        function(row) out$features[row, 1:length(lagsc)] - means[row])
      # out$features <- as.data.frame(out$features)
      # out$targets <- out$targets - means
    } else if (trend == "multiplicative") {
      means <- rowMeans(out$features)
      if (transform_features) {
        out$features <- sapply(1:nrow(out$features),
                               function(row) out$features[row, ] / means[row])
        out$features <- t(out$features)
      }
      out$features <- as.data.frame(out$features)
      out$targets <- out$targets / means
    }
  }
  if (!is.data.frame(out$features)) out$features <- as.data.frame(out$features)
  
  # Add other information to the output object
  out$call <- match.call()
  out$ts <- timeS
  out$lags <- lagsc
  out$trend <- trend
  if (trend == "differences") {
    out$nfd <- nfd
  }
  if (trend %in% c("additive", "multiplicative")) {
    out$transform_features <- transform_features
  }
  out$param <- list(...) # param
  
  # Create/train the model
  if (inherits(method, "function")) {
    # model provided by the user
    args <- c(list(X = out$features, y = out$targets), list(...))
    out$model <- do.call(method, args = args)
  } else {
    # model supported by the package
    out$model <- build_model(out$features, out$targets, method, list(...))
  }
  
  out$method <- method
  class(out) <- "utsf"
  
  out
}

#'Forecasting a time series
#'
#'@param object an object of class `utsf` embedding a forecasting model for a
#'  time series.
#'@param h A positive integer. Number of values to be forecast into the future,
#'  i.e., forecast horizon.
#'@param PI If TRUE, prediction intervals are produced using simulation and
#'  assuming normally distributed errors.
#'@param level Confidence level for predictions intervals.  
#'@param ... Other arguments passed to methods
#'
#'@returns an object of class `utsf_forecast` with the same components of the
#'  model received as first argument, plus several components:
#'  \item{`pred`}{The forecast as an `ts` object.}
#'  \item{`lower`}{Lower limits for prediction interval.}
#'  \item{`upper`}{Upper limits for prediction interval.}
#'  \item{`level`}{Confidence value associated with the prediction interval}
#' @examples
#' ## Forecast time series using k-nearest neighbors
#' m <- create_model(USAccDeaths, method = "knn")
#' f <- forecast(m, h = 12)
#' f$pred
#' library(ggplot2)
#' autoplot(f)
#'
#' ## Using k-nearest neighbors changing the default k value
#' m <- create_model(USAccDeaths, method = "knn", k = 5)
#' forecast(m, h = 12)
#'
#' ## Using your own regression model
#'
#' # Function to build the regression model
#' my_knn_model <- function(X, y, param) {
#'   structure(list(X = X, y = y), class = "my_knn")
#'}
#' # Function to predict a new example
#' predict.my_knn <- function(object, new_value) {
#'   FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
#' }
#' m <- create_model(USAccDeaths, method = my_knn_model)
#' forecast(m, h = 12)
#'@export
forecast.utsf <- function(object, h, PI = FALSE, level = 90, ...) {
  # Check h parameter
  if (! (is.numeric(h) && length(h) == 1 && h >= 1 && floor(h) == h))
    stop("h parameter should be an integer scalar greater than zero")

  # Check PI parameter
  if (! (is.logical(PI) && length(PI) == 1))
    stop("PI parameter should be a logical value")

  # Check level parameter
  if (! (is.numeric(level) && length(level) == 1 && level > 0 && level <= 100))
    stop("Confidence limit out of range")
  
  if (level < 1) level <- level * 100
  
  out <- object
  out$pred <- recursive_prediction(out, h = h)
  if (out$trend == "differences" && out$differences$differences > 0) {
    out$pred <- fd_unpreprocessing(out$pred, out$differences)
  }
  if (PI) {
    s <- try(simulations(object, h, level), silent = TRUE)
    if (inherits(s, "try-error")) {
      warning("Series too short to compute prediction intervals.")
    } else {
      out$lower <- stats::ts(s[1, ], start = stats::start(out$pred), frequency = stats::frequency(out$pred))
      out$upper <- stats::ts(s[2, ], start = stats::start(out$pred), frequency = stats::frequency(out$pred))
      out$level <- level
    }
  }
  class(out) <- "utsf_forecast"
  out
}

# @param object S3 object of class utsf
predict_one_value_transforming <- function(object, example) {
  if (object$trend == "additive") {
    mean_ <- mean(example)
    if (object$transform_features) {
      example <- example - mean_
    }
    # mean_ <- mean(head(example, length(object$lags))) # añadido
    # example[seq_along(object$lags)] <- example[seq_along(object$lags)] - mean_ # añadido
  } else if (object$trend == "multiplicative") {
    mean_ <- mean(example)
    if (object$transform_features) {
      example <- example / mean_
    }
  }
  example <- as.data.frame(matrix(example, ncol = length(example)))
  colnames(example) <- colnames(object$features)
  # print(example)
  if (inherits(object$method,"character")) {
    r <- stats::predict(object, example)
  } else {
    r <- stats::predict(object$model, example)
  }
  if (object$trend == "additive") {
    r <- r + mean_
  } else if (object$trend == "multiplicative") {
    r <- r * mean_
  }
  r
}

  

test_that("timeS parameter", {
  expect_error(create_model("a"), "timeS parameter should be of class ts or a numeric vector")
  expect_no_error(create_model(1:36, lags = 1:2))
  expect_no_error(create_model(AirPassengers))
})

test_that("lags parameter", {
  expect_error(create_model(AirPassengers, lags = "a"), "lags parameter should be NULL or numeric")
  expect_error(create_model(AirPassengers, lags = 1:3 + 0.3), "lag values should be integer")
  expect_error(create_model(AirPassengers, lags = -1:3), "lag values should be greater than zero")
  expect_equal(create_model(AirPassengers, lags = 3:1)$lags, 1:3)
})

test_that("h parameter", {
  expect_error({
      m <- create_model(AirPassengers)
      forecast(m, h = "a")
    }, "h parameter should be an integer scalar greater than zero")
  expect_error({
    m <- create_model(AirPassengers)
    forecast(m, h = 1:12)
  }, "h parameter should be an integer scalar greater than zero")
})

test_that("timeS parameter", {
  expect_error(create_model("a"), "timeS parameter should be of class ts or a numeric vector")
  expect_no_error(create_model(1:36, lags = 1:2))
  expect_no_error(create_model(AirPassengers))
})

test_that("h parameter", {
  expect_error({
    m <- create_model(AirPassengers)
    forecast(m, h = "a")
    }, "h parameter should be an integer scalar greater than zero")
  expect_error({
    m <- create_model(AirPassengers)
    forecast(m, h = 1:12)
    }, "h parameter should be an integer scalar greater than zero")
})

test_that("lags parameter", {
  expect_error(create_model(AirPassengers, lags = "a"), "lags parameter should be NULL or numeric")
})

#' Create a ggplot object from an `utsf_forecast` object
#'
#' Plot the time series and its associated forecast.
#'
#' @param object An object of class `utsf_forecast`.
#' @param ... additional parameter.
#'
#' @return The `ggplot` object representing a plotting of the time series and
#'   its forecast.
#'
#' @examples
#' m <- create_model(AirPassengers, lags = 1:12, method = "rf")
#' f <- forecast(m, h = 12)
#' library(ggplot2)
#' autoplot(f)
#' @export
#' @importFrom ggplot2 autoplot
autoplot.utsf_forecast <- function(object, ...) {
  if (is.null(object$lower)) {
    vctsfr::plot_ts(object$ts, prediction = object$pred)
  } else {
    vctsfr::plot_ts(object$ts, 
                    prediction = object$pred, 
                    lpi = object$lower,
                    upi = object$upper,
                    level = object$level
    )
  }
}

#' @importFrom graphics plot
#' @export
plot.utsf_forecast <- function(x, y, ...) {
  timeS <- stats::ts(c(x$ts, x$pred),
                     start = stats::start(x$ts),
                     frequency = stats::frequency(x$ts)
  )
  graphics::plot(timeS, type = "n", ylab = "")
  graphics::lines(x$ts, type = "o", pch = 20)
  graphics::lines(x$pred, type = "o", col = "#CC0000", pch = 20)
}

#' Build the training examples
#'
#' Build the training examples for a regressive model to forecast a time series
#' using lagged values of the series as autoregressive features.
#'
#' @param timeS The time series.
#' @param lags An integer vector with the lags used as feature vector in
#'   decreasing order.
#'
#' @return A list with two fields: 1) a matrix with the features of the examples
#'   and 2) a vector with the targets of the examples
#' @export
#'
#' @examples
#' build_examples(ts(1:5), lags = 2:1)
build_examples <- function(timeS, lags) {
  MAXLAG   <- lags[1]
  NCOL     <- length(lags)
  NROW     <- length(timeS) - MAXLAG
  features <- matrix(0, nrow = NROW, ncol = NCOL)
  targets  <- vector(mode = "numeric", length = NROW)
  row <- 1
  for (ind in seq(MAXLAG + 1, length(timeS))) {
    features[row, ] <- timeS[ind - lags]
    targets[row] <- timeS[ind]
    row <- row + 1
  }
  colnames(features) <- paste0("Lag", lags)
  list(
    features = features,
    targets = targets
  )
  # MAXLAG   <- lags[1]
  # NCOL     <- length(lags) + 1
  # NROW     <- length(timeS) - MAXLAG
  # features <- matrix(0, nrow = NROW, ncol = NCOL)
  # targets  <- vector(mode = "numeric", length = NROW)
  # row <- 1
  # for (ind in seq(MAXLAG + 1, length(timeS))) {
  #   features[row, 1:length(lags)] <- timeS[ind - lags]
  #   features[row, ncol(features)] <- stats::cycle(timeS)[ind]
  #   # features[row, ncol(features)-1] <- stats::time(timeS)[ind - lags[1]]
  #   targets[row] <- timeS[ind]
  #   row <- row + 1
  # }
  # colnames(features) <- c(paste0("Lag", lags), "Cycle")
  # list(
  #   features = features,
  #   targets = targets
  # )
}

test_that("Build examples", {
  features <- matrix(c(1:3, 2:4), nrow = 3)
  colnames(features) <- c("Lag2", "Lag1")
  r <- list(features = features, targets = 3:5)
  expect_equal(build_examples(ts(1:5), lags = 2:1), r)
})

---
title: "utsf"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{utsf}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(utsf)
```

In this document the **utsf** package is described. This package offers a meta engine for applying different regression models for univariate time series forecasting using an autoregressive approach. One main feature of the package is its extensibility, which allow you to use machine learning models not directly supported by the package, such as neural networks, or your own models. Currently, the package is mainly focused on regression tree models. You can consult an explanation of how to forecast univariate time series using regression trees here: https://arxiv.org/abs/2602.00077.

# Univariate time series forecasting ( autoregressive models) and recursive forecasts

An univariate time series forecasting method is one in which the future values of a series are predicted using only information from the series, for example, using as forecast its mean historical value. An advantage of this type of prediction is that, apart from the series being forecast, there is no need to collect any further information in order to train the forecasting model.

An autoregressive model is a kind of univariate time series forecasting model in which a value of a time series is expressed as a function of some of its past values. That is, an autoregressive model is a regression model in which the independent variables are lagged values (previous values) of the response variable. For example, given a time series with the following historical values: $t = \{1, 3, 6, 7, 9, 11, 16\}$, suppose that we want to develop an autoregressive model in which a target "is explained" by its first, second and fourth past values (in this context, a previous value is also called a *lag*, so lag 1 is the value immediately preceding a given value in the series). Given the series $t$ and lags (1, 2 and 4), the training set would be:

| Lag 4 | Lag 2 | Lag 1 | Target |
|-------|-------|-------|--------|
| 1     | 6     | 7     | 9      |
| 3     | 7     | 9     | 11     |
| 6     | 9     | 11    | 16     |

## Recursive forecasts

Given a model trained with the previous dataset, the next future value of the series is predicted as $f(Lag4, Lag2, Lag1)$, where $f$ is the regression function and $Lag4$, $Lag2$ and $Lag1$ are the fourth, second and first lagged values of the next future value. So, the next future value of series $t$ is predicted as $f(7, 11, 16)$, producing a value that will be called $F1$.

Suppose that the *forecast horizon* (the number of future values to be forecast into the future) is greater than 1. In the case that the regression function only predicts the next future value of the series, a recursive approach can be applied to forecast all the future values in the forecast horizon. Using a recursive approach, the regression function is applied recursively until all steps ahead values are forecast. For instance, following the previous example, suppose that the forecast horizon is 3. As we have explained, to forecast the next future value of the series (one-step ahead) the regression function is fed with the vector $[7, 11, 16]$, producing $F1$. To forecast the two-steps ahead value the regression function is fed with the vector $[9, 16, F1]$. The forecast for the one-step ahead value, $F1$, is used as the first lag for the two-steps ahead value, because the actual value is unknown. Finally, to predict the three-steps ahead value the regression function is fed with the vector [$11, F1, F2]$. This example of recursive forecast is summarized in the following table:

|Steps ahead | Autoregressive values | Forecast |
|------------|-----------------------|----------|
| 1          | 7, 11, 16             | F1       |
| 2          | 9, 16, F1             | F2       |
| 3          | 11, F1, F2            | F3       |

The recursive approach for forecasting several values into the future is applied in classical statistical models such as ARIMA or exponential smoothing.

# The utsf package

The **utsf** package makes it easy the use of classical regression models for univariate time series forecasting, employing the autoregressive approach and the recursive prediction strategy explained in the previous section. All the supported models are applied using an uniform interface: 

* the `create_model()` function to build the forecasting model, and 
* the `forecast()` function for using the trained model to predict future values. 

Let us see an example in which a regression tree model is used to forecast the next future values of a time series:

```{r}
m <- create_model(AirPassengers, lags = 1:12, method = "rt")
f <- forecast(m, h = 12)
```

In this example, an autoregressive tree model (`method = "rt"`, Regression Tree) is trained using the historical values of the `AirPassengers` time series and a forecast for its 12 next future values (`h = 12`) is done. The `create_model()` function returns an S3 object of class `utsf` with information about the trained model. The information about the forecast is included in the object returned by the `forecast()` function, as a component named `pred` of class `ts` (a time series):

```{r}
f$pred
library(ggplot2)
autoplot(f)
```

The training set used to fit the model is built from the historical values of the time series using the autoregressive approach explained in the previous section. The `lags` parameter of the `create_model()` function is used to specify the autoregressive lags. In the example: `lags = 1:12`, so a target is a function of its 12 previous values. Next, we consult the first targets (and their associated features) with which the regression model has been trained:

```{r}
head(m$targets)  # first targets
head(m$features) # and its associated features
```

The curious reader might have noticed that the features and targets are not on the same scale as
the original time series. This is because, by default, a transformation is applied to the examples of the training set. This transformation will be explained later.

Let us see the training set associated with the example of the previous section:

```{r}
t <- ts(c(1, 3, 6, 7, 9, 11, 16))
out <- create_model(t, lags = c(1, 2, 4), trend = "none")
cbind(out$features, Target = out$targets)
```

No transformation has been applied this time (`trend = "none"`).

# Prediction intervals

Prediction intervals for a forecast can be computed:

```{r}
m <- create_model(USAccDeaths, method = "mt")
f <- forecast(m, h = 12, PI = TRUE, level = 90)
f
library(ggplot2)
autoplot(f)
```

Prediction intervals are calculated following the guidelines in (https://otexts.com/fpp3/nnetar.html#prediction-intervals-5). Random errors are assumed to follow a normal distribution. In the example, a 90% prediction interval (`level = 90`) has been computed.

# Supported models

The `create_model()` and `forecast()` functions provide a common interface to applying an autoregressive approach for time series forecasting using different regression models. These models are implemented in several R packages. Currently, our project is mainly focused on regression tree models, supporting the following approaches:

* k-nearest neighbors: In this case no model is trained and the function `FNN::knn.reg()` is used, as regression function, to recursively predict the future values of the time series.
* Linear models: The model is trained using the function `stats::lm()` and its associated method
`stats::predict.lm()` is applied recursively for the forecasts, i.e., as regression function.
* Regression trees: The model is trained using the function `rpart::rpart()` and its associated method `rpart::predict.rpart()` is used for the forecasts.
* Model trees: The model is trained with the function `Cubist::cubist()` and its associated method `Cubist::predict.cubist()` is used for predictions.
* Bagging: The model is trained with the function `ipred::bagging()` and its
  associated method `ipred::predict.regbagg()` is used for forecasting.
* Random forest: The model is trained with the function `ranger::ranger()` and its associated method `ranger::predict.ranger()` is used for predictions.
* Extreme gradient boosting: The model is trained with the function `xgboost::xgboost()` and its associated method `xgboost::predict.xgboost()` is used for predictions.

The S3 object of class `utsf` returned by the `create_model()` function contains a component with the trained autoregressive model:

```{r}
m <- create_model(fdeaths, lags = 1:12, method = "rt")
m$model
```

In this case, the model is the result of training a regression tree  using the function `rpart::rpart()` with the training set consisting of the features `m$features` and targets `m$targets`. Once the model is trained, the `rpart::predict.rpart()` function can be used recursively to forecast the future values of the time series using the `forecast()` function.

# Using your own models

An interesting feature of the **utsf** package is its extensibility. Apart from the models directly supported by the package, you can use the facilities of the package to do autoregressive time series forecasting using your own regression models. Thus, your models can benefit from the features implemented in the package, such as the building of the training set, the implementation of recursive forecasts, pre-processings, the estimation of forecast accuracy or parameter tuning.

To apply your own regression model you have to use the `method` parameter of the `create_model()` function, providing as argument a function that is able to train your model. We will call this function the *modeling function* and it must return an object with the trained regression model. The modeling function must have three input parameters:

* `X`: it is a data frame with the features of the training examples. This data frame is built from the time series taking into account the autoregressive lags as explained in a previous section. This is the same object as the `features` component of the object returned by the `create_model()` function.
* `y`: a vector with the targets of the training examples. It is built as explained in a previous section. It is the same object as the `targets` component of the object returned by the `create_model()` function.
* `...`: for receiving additional arguments to adjust the behavior of the model. This parameter will not be used in this section and will be explained in the next section.

Furthermore, if the modeling function returns a model of class `model_class`, a method with the signature `predict.model_class(object, new_value)` should be implemented. This method uses your model to predict a new value, that is, it is the regression function associated with the model. Let us explain the parameters of this regression function:

* `object`: it is the object of class `model_class` returned by your function that creates the model.
* `new_value`: it is a data frame with the same structure as the `X` parameter of the function for building the model. The `new_value` data frame has only one row, with the features of the example to be predicted. 

Summing up, in order to use your regression model you have to provide:

* A modeling function that, given the training set, trains the model.
* A regression function that, given a trained model and a new vector of features, predicts a value.

Let us see several examples of how to use this functionality for applying your own models for autoregressive time series forecasting taking advantage of the capabilities of the **utsf** package.

## Example 1: k-nearest neighbors

The case of K-nearest neighbors is a bit special because it is a lazy learner and no model is built. You only need to store the training set and all the work is done by the regression function. Let us see a first example in which we rely on the implementation of k-NN in the **FNN** package:

```{r}
# Modeling function: in this case (k-NN) just stores the training set
my_knn_model <- function(X, y, ...) { structure(list(X = X, y = y), class = "my_knn") }

# Regression function
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
}

m <- create_model(AirPassengers, lags = 1:12, method = my_knn_model)
f <- forecast(m, h = 12)
f$pred
autoplot(f)
```

The modeling function (`my_knn_model()`) creates a "model" that only stores the training set. The regression function (`predict.my_knn()`) takes advantage of the `FNN::knn.reg()` function to look for the k-nearest neighbors and compute their mean response value. The default number of neighbors (3) of `FNN::knn.reg()` is used. Later, we will modify this example so that the user can select the $k$ value.

The k-nearest neighbors algorithm is so simple that it can be easily implemented without using functionality from any R package:

```{r}
# Modeling function
my_knn_model2 <- function(X, y, ...) {
  structure(list(X = X, y = y), class = "my_knn2")
}

# Regression function
predict.my_knn2 <- function(object, new_value) {
  k <- 3 # number of nearest neighbors
  distances <- sapply(1:nrow(object$X), function(i) sum((object$X[i, ] - new_value)^2))
  k_nearest <- order(distances)[1:k]
  mean(object$y[k_nearest])
}

m2 <- create_model(AirPassengers, lags = 1:12, method = my_knn_model2)
forecast(m2, h = 12)$pred
```
## Example 2: random forest and neural networks

The random forest algorithm is directly supported in the **utsf** package, using the implementation of random forests in the `ranger` package. Here, we are going to create an autoregressive model based on the random forest model implemented in the `RandomForest` package:

```{r}
library(randomForest)
my_model <- function(X, y, ...) { randomForest(x = X, y = y) }
m <- create_model(USAccDeaths, lags = 1:12, method = my_model)
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)
print(m$model)
```

The modeling function (`my_model()`) just uses the `randomForest::randomForest()` function. This function returns an S3 object of class `randomForest` with the trained model. In this case, we have not implemented the regression function, because the `predict.randomForest()` method does exactly what we want.

As another example, we are going to use a neural network model implemented in the *nnet* package:

```{r}
library(nnet)
my_model <- function(X, y, ...) {
  nnet(x = X, y = y, size = 5, linout = TRUE, trace = FALSE)
}
m <- create_model(USAccDeaths, lags = 1:12, method = my_model)
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)
```

In this case, the `nnet()` function returns an S3 object of class `nnet`. Again, the regression method associated with this class, `predict.nnet()`,  is what we need as regression function. Because we are using neural networks, probably some pre-processing would be needed to obtain more accurate predictions. 

# Setting the parameters of the regression models

Normally, a regression model can be adjusted using different parameters. By default, the models supported by our package are set using some specific parameters, usually the default values of the functions used to train the models (these functions are listed in a previous section). However, the user can select these parameters. Let us see an example:

```{r}
# A bagging model set with default parameters
m <- create_model(USAccDeaths, lags = 1:12, method = "bagging")
length(m$model$mtrees) # number of regression trees (25 by default)
# A bagging model set with 3 regression trees
m2 <- create_model(USAccDeaths,
                   lags = 1:12,
                   method = "bagging",
                   nbagg = 3
)
length(m2$model$mtrees) # number of regression trees
```

In the previous example, two bagging models (using regression trees) are trained with the `create_model()` function. In the first model the number of trees is 25, the default value of the function `ipred::ipredbagg()` used to train the model. In the second model the number of trees is set to 3. Of course, in order to set some specific parameters the user must consult the parameters of the function used internally by the **utsf** package to train the model. In the example, the `ipred::ipredbagg()` function has a parameter named `nbagg` to set the number of regression trees.

## Setting the paramaters of a regression model supplied by the user

When the regression model is provided by the user, it is also possible to adjust its parameters. As explained previously, if you want to use your own regression model you have to provide a modeling function. This function has tree parameters:

* `X`: to receive the training features
* `y`: to receive the targets
* `...`: additional arguments for adjusting the model. 

The `...` parameter allows the modeling function to receive a variable name of arguments for adjusting the model.

Let us see some example of passing arguments to some of the regression models implemented in the previous section.

### Example 1: k-nearest neighbors

In this case the model can be adjusted specifying the $k$ value. The function provided to train the model can be tuned with an argument named `k` with the number of nearest neighbors:

```{r}
# Function to train the model
my_knn_model <- function(X, y, ...) {
  param <- list(...)
  k <- if ("k" %in% names(param)) param$k else 3
  structure(list(X = X, y = y, k = k), class = "my_knn")
}

# Regression function for object of class my_knn
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value,
               y = object$y, k = object$k)$pred
}

# The model is trained with default parameters (k = 3)
m <- create_model(AirPassengers, lags = 1:12,  method = my_knn_model)
print(m$model$k)
# The model is trained with k = 5
m2 <- create_model(AirPassengers, method = my_knn_model, k = 5)
print(m2$model$k)
```

The additional arguments set in the `create_model()` function (in this case, `k = 5`) are passed to the modeling function (in this case, `my_knn_model()`).

### Example 2: random forest

As another example, suppose that we create a random forest model using the functionality in the *randomForest* package. By default, we will use the default parameters of the `randomForest::randomForest()` function to build the model, save for the number of trees (it will be set to 200). However, the user can train the model with other values:

```{r}
library(randomForest)
my_model <- function(X, y, ...) {
    param <- list(...)
    args <- list(x = X, y = y, ntree = 200) # default parameters
    args <- args[!(names(args) %in% names(param))]
    args <- c(args, param)
    do.call(randomForest::randomForest, args = args)
}
# The random forest is built with our default parameters
m <- create_model(USAccDeaths, method = my_model)
print(m$model$ntree)
print(m$model$mtry)
# The random forest is built with ntree = 400 and mtry = 6
m <- create_model(USAccDeaths, method = my_model, ntree = 400, mtry = 6)
print(m$model$ntree)
print(m$model$mtry)
```

# Estimating forecast accuracy

This section explains how to estimate the forecast accuracy of a regression model predicting a time series with the **utsf** package. Let us see an example:

```{r}
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- efa(m, h = 4, type = "normal", size = 8)
r$per_horizon
r$global
```

To assess the forecast accuracy of a model you should use the `efa()` function on the model. In this case, the forecasting method is a K-nearest neighbors algorithm (`method = "knn"`), with autoregressive lags 1 to 4, applied on the `UKgas` time series. An estimation of its forecast accuracy on the series for a forecasting horizon of 4 (`h = 4`) is obtained according to different forecast accuracy measures. The `efa()` function returns a list with two components. The component named `per_horizon` contains the forecast accuracy estimations per forecasting horizon. The component named `global` is computed as the means of the rows of the `per_horizon` component. Currently, the following forecast accuracy measures are computed:

* MAE: mean absolute error
* MAPE: mean absolute percentage error
* sMAPE: symmetric MAPE
* RMSE: root mean squared error

Next, we describe how the forecasting accuracy measures are computed for a forecasting horizon $h$ ($y_t$ and $\hat{y}_t$ are the actual future value and its forecast for horizon $t$ respectively):

$$
MAE = \frac{1}{h}\sum_{t=1}^{h} |y_t-\hat{y}_t|
$$

$$
MAPE = \frac{1}{h}\sum_{t=1}^{h} 100\frac{|y_t-\hat{y}_t|}{y_t}
$$
$$
sMAPE = \frac{1}{h}\sum_{t=1}^{h} 200\frac{\left|y_{t}-\hat{y}_{t}\right|}{|y_t|+|\hat{y}_t|}
$$

$$
RMSE = \sqrt{\frac{1}{h}\sum_{t=1}^{h} (y_t-\hat{y}_t)^2}
$$

The estimation of forecast accuracy is done with the well-known rolling origin evaluation. This technique builds several training and test sets from the series, as shown below for a forecasting horizon of 4:

```{r out.width = '85%', echo = FALSE}
knitr::include_graphics("ro_normal.png")
```

In this case, the last nine observations of the series are used to build the test sets. Six models are trained using the different training sets and their forecasts are compared to their corresponding test sets. Thus, the estimation of the forecast error for every forecasting horizon is based on six errors (it is computed as the mean).

You can specify how many of the last observations of the series are used to build the test sets with the `size` and `prop` parameters. For example:

```{r, eval = FALSE}
m <- create_model(UKgas, lags = 1:4, method = "knn")
# Use the last 9 observations of the series to build test sets
r1 <- efa(m, h = 4, type = "normal", size = 9)
# Use the last 20% observations of the series to build test sets
r2 <- efa(m, h = 4, type = "normal", prop = 0.2)
```

If the series is short, the training sets might be too small to fit a meaningful model. In that case, a special rolling origin evaluation can be done, setting parameter `type` to `"minimum"`:

```{r}
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- efa(m, h = 4, type = "minimum")
r$per_horizon
r$global
```

When this option is chosen, the last `h` observations (in the example 4) are used to build the training sets as is shown below for `h = 4`:

```{r out.width = '85%', echo = FALSE}
knitr::include_graphics("ro_minimum.png")
```

In this case, the estimated forecast error for horizon 1 is based on 4 errors, for horizon 2 is based on 3 errors, for horizon 3 on 2 errors and for horizon 4 on 1 error.

The `efa()` function also returns the test sets (and the predictions associated with them) used when assessing the forecast accuracy:

```{r}
timeS <- ts(1:25)
m <- create_model(timeS, lags = 1:3, method = "mt")
r <- efa(m, h = 5, size = 7)
r$test_sets
r$predictions
```

Each row in these tables represent a test set or prediction. Let us see the test sets when the "minimum" evaluation is done:

```{r}
timeS <- ts(1:25)
m <- create_model(timeS, lags = 1:3, method = "mt")
r <- efa(m, h = 3, type = "minimum")
r$test_sets
r$predictions
```

# Parameter tuning

Another useful feature of the **utsf** package is parameter tuning. The `tune_grid()` function allows you to estimate the forecast accuracy of a model using different combinations of parameters. Furthermore, the combination of parameters that gets the best forecast accuracy estimation is used to train a model with all the historical values of the series and the model is applied for forecasting the future values of the series. Let us see an example:

```{r}
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- tune_grid(m, h = 4, tuneGrid = expand.grid(k = 1:7), type = "normal", size = 8)
# To see the estimated forecast accuracy with the different configurations
r$tuneGrid
```

In this example, the `tuneGrid` parameter is used to specify (using a data frame) the set of parameters to assess. The forecast accuracy of the model for the different combinations of parameters is estimated as explained in the previous section using the last observations of the time series as validation set. The `size` parameter is used to specify the size of the test set. The `tuneGrid` component of the list returned by the `tune_grid()` function contains the result of the estimation. In this case, the k-nearest neighbors method using $k=1$ obtains the best results for all the forecast accuracy measures (this is due to the increasing variance of the series). The best combination of parameters according to RMSE is used to forecast the time series:

```{r}
r$best
r$forecast
```

Let us plot the values of $k$ against their estimated accuracy using RMSE:

```{r}
plot(r$tuneGrid$k, r$tuneGrid$RMSE, type = "o", pch = 19, xlab = "k (number of nearest neighbors)", ylab = "RMSE", main = "Estimated accuracy")
```

Let us now see an example in which more than one tuning parameter is used. In this case we use a random forest model and three tuning parameters:

```{r}
m <- create_model(UKDriverDeaths, lags = 1:12, method = "rf")
r <- tune_grid(m, 
               h = 12, 
               tuneGrid = expand.grid(num.trees = c(200, 500), replace = c(TRUE, FALSE), mtry = c(4, 8)),
               type = "normal", 
               size = 12
)
r$tuneGrid
```

# Preprocessings and transformations

Sometimes, the forecast accuracy of a model can be improved by applying some transformations and/or preprocessings to the time series being forecast. Currently, the **utsf** package is focused on preprocessings related to forecasting time series with a trend. This is due to the fact that, at present, most of the models incorporated in our package (such as regression trees and k-nearest neighbors) predict averages of the targets in the training set (i.e., an average of historical values of the series). In a trended series these averages are probably outside the range of the future values of the series. Let us see an example in which a trended series (the yearly revenue passenger miles flown by commercial airlines in the United States) is forecast using a random forest model:

```{r}
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "none")
f <- forecast(m, h = 4)
autoplot(f)
```

In this case, no preprocessing is applied for dealing with the trend (`trend = "none"`). It can be observed how the forecast does not capture the trending behavior of the series because regression tree models predict averaged values of the training targets.

In the next subsections we explain how to deal with trended series using three different transformations.

## Differencing

A common way of dealing with trended series is to transform the original series taking first differences (computing the differences between consecutive observations). Then, the model is trained with the differenced series and the forecasts are back-transformed. Let us see an example:

```{r}
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "differences", nfd = 1)
f <- forecast(m, h = 4)
autoplot(f)
```

In this case, the `trend` parameter has been used to specify first-order differences (`trend = "differences"`). The order of first differences is specified with the `nfd` parameter (it will be normally 1). It is also possible to specify the value -1, in which case the order of differences is estimated using the `ndiffs()` function from the **forecast** package (in that case, the chosen order of first differences could be 0, that is, no differences are applied).

```{r}
# The order of first differences is estimated using the ndiffs function
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "differences", nfd = -1)
m$differences
```

## The additive transformation

This transformation has been used to deal with trending series in other R packages, such as **tsfknn** or **tsfgrnn**. The additive transformation works transforming the targets of the training set as follows:

* The target associated with a vector of features is transformed by subtracting from it the mean of its vector of features.
* To back transform a prediction, the mean of the input vector is added to it.

It is easy to check how the training examples are modified by the additive transformation using the API of the package. For example, let us see the training examples of a model with no transformation:

```{r}
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "none")
cbind(m$features, Targets = m$targets)
```

Now, let us see the effect of the additive transformation in the training examples:

```{r}
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "additive", transform_features = FALSE)
cbind(m$features, Targets = m$targets)
```

Apart from transforming the targets, it is also possible to transform the features. A vector of features is transformed by substracting the mean of the vector. Let us see the effects of this transformation:

```{r}
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "additive", transform_features = TRUE)
cbind(m$features, Targets = m$targets)
```

The point of transforming the features is to remove the effects of the level of the series in the feature space. Our experience tells us that transforming the features tends to improve forecast accuracy. 

Next, we forecast the airmiles series using the additive transformation and a random forest model:

```{r}
m <- create_model(airmiles, lags = 1:4, method = "rf")
f <- forecast(m, h = 4)
autoplot(f)
```

By default, the additive transformation is applied to both targets and features.

## The multiplicative transformation

The multiplicative transformation is similar to the additive transformation, but it is intended for series with an exponential trend (the additive transformation is suited for series with a linear trend). In the multiplicative transformation a training example is transformed this way:

* The target associated with a vector of features is transformed by dividing it by the mean of its associated vector of features.
* To back transform a prediction, the prediction is multiplied by the mean of the input vector.
* Optionally, a vector of features is transformed by dividing it by its mean.

Let us see an example of an artificial time series with a multiplicative trend and its forecast using the additive and the multiplicative transformation:

```{r}
t <- ts(10 * 1.05^(1:20))
m_m <- create_model(t, lags = 1:3, method = "rf", trend = "multiplicative")
f_m <- forecast(m_m, h = 4)
m_a <- create_model(t, lags = 1:3, method = "rf", trend = "additive")
f_a <- forecast(m_a, h = 4)
library(vctsfr)
plot_predictions(t, predictions = list(Multiplicative = f_m$pred, Additive = f_a$pred))
```

In this case, the forecast with the multiplicative transformation captures perfectly the exponential trend.

As another example, let us compare the additive and multiplicative transformation applied to the `AirPassengers` series. This monthly series has an exponential trend.

```{r, fig.width= 6}
m_m <- create_model(AirPassengers, lags = 1:12, method = "knn", trend = "multiplicative")
f_m <- forecast(m_m, h = 36)
m_a <- create_model(AirPassengers, lags = 1:12, method = "knn", trend = "additive")
f_a <- forecast(m_a, h = 36)
library(vctsfr)
plot_predictions(AirPassengers, predictions = list(Multiplicative = f_m$pred, Additive = f_a$pred))
```



# Default parameters

The `create_model()` function only has one compulsory parameter: the time series with which the model is trained. Next, we discuss the default values for the rest of its parameters:

* `lags`: The `lags` parameter is an integer vector with the autoregressive lags. If `frequency(ts) == f` where `ts` is the time series being forecast and $f > 1$ then the lags used as autoregressive features are 1:*f*. For example, the lags for quarterly data are 1:4 and for monthly data 1:12. This is useful to capture a possible seasonal behavior of the series. If `frequency(ts) == 1`, then:
    * The lags with significant autocorrelation in the partial autocorrelation function (`stats::pacf()`) are selected.
    * If no lag has a significant autocorrelation, then lags 1:5 are chosen.
    * If only one lag has significant autocorrelation and the additive or multiplicative transformation is applied to both the targets and the features, then lags 1:5 are chosen. This is done because it does not make sense to use these transformations with the features when only one autoregressive lag is used.
* `method`: By default, the K-nearest neighbors algorithm is applied.
* `trend`: The additive transformation is applied. This transformation have proved its effectiveness to forecast trending series. In general, its application seems beneficial on any kind of series.
* `transform_features`: Its default value is `TRUE`, so that if a series is transformed with the additive or multiplicative transformation both its features and targets are transformed.

---
title: "utsf"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{utsf}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(utsf)
```

In this document, the **utsf** package is described. This package offers a meta-engine for applying different regression models to univariate time series forecasting using an autoregressive approach. One of the main features of the package is its extensibility, which allows you to use machine learning models not directly supported by the package (such as neural networks) or your own custom models. Currently, the package is mainly focused on regression tree models. An explanation of how to forecast univariate time series using regression trees can be found here: <https://arxiv.org/abs/2602.00077>.

# Univariate time series forecasting (autoregressive models) and recursive forecasts

A univariate time series forecasting method is one in which the future values of a series are predicted using only information from the series itself—for example, using its historical mean value as a forecast. An advantage of this type of prediction is that, apart from the series being forecasted, there is no need to collect any further information to train the forecasting model.

An autoregressive model is a type of univariate time series forecasting model in which a value of a time series is expressed as a function of some of its past values. That is, an autoregressive model is a regression model in which the independent variables are lagged values (previous values) of the response variable. For example, given a time series with the following historical values: $t = \{1, 3, 6, 7, 9, 11, 16\}$, suppose that we want to develop an autoregressive model in which a target "is explained" by its first, second, and fourth past values (in this context, a previous value is also called a *lag*, so lag 1 is the value immediately preceding a given value in the series). Given the series $t$ and lags (1, 2, and 4), the training set would be:

| Lag 4 | Lag 2 | Lag 1 | Target |
|-------|-------|-------|--------|
| 1     | 6     | 7     | 9      |
| 3     | 7     | 9     | 11     |
| 6     | 9     | 11    | 16     |

## Recursive forecasts

Given a model trained with the previous dataset, the next future value of the series is predicted as $f(\text{Lag}_4, \text{Lag}_2, \text{Lag}_1)$, where $f$ is the regression function and $\text{Lag}_4$, $\text{Lag}_2$, and $\text{Lag}_1$ are the fourth, second, and first lagged values of the next future value. Therefore, the next future value of the series $t$ is predicted as $f(7, 11, 16)$, producing a value that will be denoted as $F_1$.

Suppose that the *forecast horizon* (the number of future values to be forecasted into the future) is greater than 1. If the regression function only predicts the next future value of the series, a recursive approach can be applied to forecast all the values within the horizon. Using this approach, the regression function is applied recursively until all multi-step-ahead values are forecasted. For instance, following the previous example, suppose that the forecast horizon is 3. As explained, to forecast the next value of the series (one-step-ahead) the regression function is fed with the vector $[7, 11, 16]$, producing $F_1$. To forecast the two-step-ahead value the regression function is fed with the vector $[9, 16, F_1]$. The forecast for the one-step-ahead value, $F_1$, is used as the first lag for the two-step-ahead forecast, because the actual value is unknown. Finally, to predict the three-step-ahead value, the regression function is fed with the vector [$11, F_1, F_2]$. This example of recursive forecast is summarized in the table below:

|Steps ahead | Autoregressive values | Forecast |
|------------|-----------------------|----------|
| 1          | 7, 11, 16             | $F_1$    |
| 2          | 9, 16, $F_1$          | $F_2$    |
| 3          | 11, $F_1$, $F_2$      | $F_3$    |

The recursive approach to multi-step-ahead forecasting is employed in classical statistical models, such as ARIMA and exponential smoothing.

# The utsf package

The **utsf** package simplifies the use of classical regression models for univariate time series forecasting, employing the autoregressive approach and the recursive prediction strategy explained in the previous section. All the supported models are applied using a uniform interface: 

* the `create_model()` function to build the forecasting model, and 
* the `forecast()` function for using the trained model to predict future values. 

Let us see an example in which a regression tree model is used to forecast the future values of a time series:

```{r}
m <- create_model(AirPassengers, lags = 1:12, method = "rt")
f <- forecast(m, h = 12)
```

In this example, an autoregressive tree model (`method = "rt"`, Regression Tree) is trained using the historical values of the `AirPassengers` time series and a forecast for its next 12 values (`h = 12`) is generated. The `create_model()` function returns an S3 object of class `utsf` with information about the trained model. The forecast information is included in the object returned by the `forecast()` function, as a component named `pred` of class `ts` (a time series):

```{r}
f$pred
library(ggplot2)
autoplot(f)
```

The training set used to fit the model is built from the historical values of the time series using the autoregressive approach explained in the previous section. The `lags` parameter of the `create_model()` function specifies the autoregressive lags. In this example: `lags = 1:12`, meaning each target is a function of its 12 previous values. Next, we check the first targets (and their associated features) used to train the regression model:

```{r}
head(m$targets)  # first targets
head(m$features) # and its associated features
```

The curious reader may have noticed that the features and targets are not on the same scale as the original time series. This is because, by default, a transformation is applied to the examples of the training set. This transformation will be explained later.

Let us inspect the training set associated with the example from the previous section:

```{r}
t <- ts(c(1, 3, 6, 7, 9, 11, 16))
out <- create_model(t, lags = c(1, 2, 4), trend = "none")
cbind(out$features, Target = out$targets)
```

No transformation has been applied this time (`trend = "none"`).

# Prediction intervals

Prediction intervals for a forecast can be computed:

```{r}
m <- create_model(USAccDeaths, method = "mt")
f <- forecast(m, h = 12, PI = TRUE, level = 90)
f
library(ggplot2)
autoplot(f)
```

Prediction intervals are calculated following the guidelines in [Forecasting: Principles and Practice](https://otexts.com/fpp3/nnetar.html#prediction-intervals-5). Random errors are assumed to follow a normal distribution. In this example, a 90% prediction interval (`level = 90`) was computed.

# Supported models

The `create_model()` and `forecast()` functions provide a unified interface for applying an autoregressive approach to time series forecasting with various regression models. These models are implemented across several R packages. Currently, our project focuses primarily on regression tree models, supporting the following approaches:

* k-nearest neighbors: In this case no model is trained and the function `FNN::knn.reg()` is used, as regression function, to recursively predict the future values of the time series.
* Linear models: The model is trained using the function `stats::lm()` and its associated method
`stats::predict.lm()` is applied recursively for the forecasts, i.e., as regression function.
* Regression trees: The model is trained using the function `rpart::rpart()` and its associated method `rpart::predict.rpart()` is used for the forecasts.
* Model trees: The model is trained with the function `Cubist::cubist()` and its associated method `Cubist::predict.cubist()` is used for predictions.
* Bagging: The model is trained with the function `ipred::bagging()` and its
  associated method `ipred::predict.regbagg()` is used for forecasting.
* Random forest: The model is trained with the function `ranger::ranger()` and its associated method `ranger::predict.ranger()` is used for predictions.
* Extreme gradient boosting: The model is trained with the function `xgboost::xgboost()` and its associated method `xgboost::predict.xgboost()` is used for predictions.

The S3 object of class `utsf` returned by the `create_model()` function contains a component with the trained autoregressive model:

```{r}
m <- create_model(fdeaths, lags = 1:12, method = "rt")
m$model
```

In this case, the model is obtained by training a regression tree using the `rpart::rpart()` function on a training set composed of the features `m$features` and targets `m$targets`. Once trained, the `rpart::predict.rpart()` function can be used recursively within `forecast()` to predict future values of the time series.

# Using your own models

An appealing feature of the **utsf** package is its extensibility. Beyond the natively supported models, you can leverage the package's infrastructure to perform autoregressive time series forecasting with custom regression models. This allows user-defined models to benefit from built-in functionality, such as training set construction, recursive forecasting, prediction intervals, preprocessing, accuracy evaluation, and hyperparameter tuning.

To use a custom regression model, pass a function capable of training your model to the `method` parameter of the `create_model()` function. We refer to this function as the *modeling function*, and it must return an object containing the trained regression model. The modeling function must accept three input parameters:

* `X`: a data frame containing the features of the training examples. This data frame is constructed from the time series according to the autoregressive lags, as explained in a previous section. It is identical to the `features` component of the object returned by `create_model()`.

* `y`: a vector containing the targets of the training examples, constructed as described in a previous section. It is identical to the `targets` component of the object returned by `create_model()`.
* `...`: additional arguments passed to fine-tune model behavior. This parameter is not used in this section and will be detailed in the next.

Furthermore, if the modeling function returns an object of class `model_class`, an S3 method with the signature `predict.model_class(object, new_value)` must be implemented. This method uses the model to predict a single new value (acting as the prediction function associated with the model). Its parameters are described below:

* `object`: the object of class `model_class` returned by your modeling function.
* `new_value`: a one-row data frame with the same structure as `X`, containing the features of the instance to be predicted.

In summary, to use a custom regression model, you must provide:

* A modeling function that trains the model given a training set.
* A prediction function that predicts a single target value given a trained model and a new feature vector.

The following examples demonstrate how to use this functionality to apply custom models to autoregressive time series forecasting, leveraging the capabilities of the **utsf** package.

## Example 1: k-nearest neighbors

The k-nearest neighbors (k-NN) algorithm is a special case because it is a lazy learner and builds no explicit model. Instead, it simply stores the training set, leaving all the computation to the prediction function. Below is an initial example using the k-NN implementation from the **FNN** package:

```{r}
# Modeling function: in this case (k-NN) just stores the training set
my_knn_model <- function(X, y, ...) { structure(list(X = X, y = y), class = "my_knn") }

# Regression function
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
}

m <- create_model(AirPassengers, lags = 1:12, method = my_knn_model)
f <- forecast(m, h = 12)
f$pred
autoplot(f)
```

The modeling function (`my_knn_model()`) creates a "model" that simply stores the training set. The regression function (`predict.my_knn()`) uses `FNN::knn.reg()` to find the $k$-nearest neighbors and compute their mean response value. The default number of neighbors in `FNN::knn.reg()` (3) is used. Later, we will modify this example to allow users to specify the value of $k$.

The $k$-nearest neighbors algorithm is simple enough to be easily implemented from scratch, without relying on external R packages:

```{r}
# Modeling function
my_knn_model2 <- function(X, y, ...) {
  structure(list(X = X, y = y), class = "my_knn2")
}

# Regression function
predict.my_knn2 <- function(object, new_value) {
  k <- 3 # number of nearest neighbors
  distances <- sapply(1:nrow(object$X), function(i) sum((object$X[i, ] - new_value)^2))
  k_nearest <- order(distances)[1:k]
  mean(object$y[k_nearest])
}

m2 <- create_model(AirPassengers, lags = 1:12, method = my_knn_model2)
forecast(m2, h = 12)$pred
```
## Example 2: random forest and neural networks

The random forest algorithm is natively supported in the **utsf** package via the `ranger` package implementation. Here, we build an autoregressive model using the random forest implementation provided by the `randomForest` package:

```{r}
library(randomForest)
my_model <- function(X, y, ...) { randomForest(x = X, y = y) }
m <- create_model(USAccDeaths, lags = 1:12, method = my_model)
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)
print(m$model)
```

The modeling function (`my_model()`) simply calls `randomForest::randomForest()`, which returns an S3 object of class `randomForest` containing the trained model. In this case, we do not need to implement a regression function, as the existing `predict.randomForest()` S3 method already performs the required task.

As another example, we implement an autoregressive neural network model using the **nnet** package:

```{r}
library(nnet)
my_model <- function(X, y, ...) {
  nnet(x = X, y = y, size = 5, linout = TRUE, trace = FALSE)
}
m <- create_model(USAccDeaths, lags = 1:12, method = my_model)
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)
```

In this case, `nnet()` returns an S3 object of class `nnet`. Once again, the class's default prediction method, `predict.nnet()`, provides the required functionality. Given that neural networks are sensitive to feature scales, preprocessing the time series would likely improve forecast accuracy.

# Setting the parameters of the regression models

Typically, a regression model can be tuned using hyperparameters. By default, the models supported by the package use specific default values defined by their underlying training functions (listed in a previous section). However, users can customize these hyperparameters to adjust model behavior. Below is an example:

```{r}
# A bagging model set with default parameters
m <- create_model(USAccDeaths, lags = 1:12, method = "bagging")
length(m$model$mtrees) # number of regression trees (25 by default)
# A bagging model set with 3 regression trees
m2 <- create_model(USAccDeaths,
                   lags = 1:12,
                   method = "bagging",
                   nbagg = 3
)
length(m2$model$mtrees) # number of regression trees
```

In the previous example, two bagging models (using regression trees) are trained with `create_model()`. The first model uses 25 trees (the default for `ipred::ipredbagg()`) while the second sets the number of trees to 3. To customize hyperparameters, consult the parameter documentation of the training function used internally by **utsf**. In this case, `ipred::ipredbagg()` uses the `nbagg` parameter to specify the number of trees in the ensemble.

## Setting the paramaters of a regression model supplied by the user

When using a custom regression model, its parameters can also be tuned. As explained previously, custom models require a modeling function with three parameters:

* `X`: a data frame containing the training features.
* `y`: a vector containing the target values.
* `...`: additional arguments passed to adjust model behavior.

The `...` parameter allows the modeling function to accept an arbitrary number of arguments to tune model behavior.

Below are several examples of passing arguments to the custom regression models implemented in the previous section:

### Example 1: k-nearest neighbors

In this case, the model can be tuned by specifying the value of $k$. The modeling function accepts a parameter named `k` to set the number of nearest neighbors:

```{r}
# Function to train the model
my_knn_model <- function(X, y, ...) {
  param <- list(...)
  k <- if ("k" %in% names(param)) param$k else 3
  structure(list(X = X, y = y, k = k), class = "my_knn")
}

# Regression function for object of class my_knn
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value,
               y = object$y, k = object$k)$pred
}

# The model is trained with default parameters (k = 3)
m <- create_model(AirPassengers, lags = 1:12,  method = my_knn_model)
print(m$model$k)
# The model is trained with k = 5
m2 <- create_model(AirPassengers, method = my_knn_model, k = 5)
print(m2$model$k)
```

Additional arguments specified in `create_model()` (such as `k = 5`) are passed directly to the custom modeling function (`my_knn_model()`).

### Example 2: random forest

As another example, suppose we create a random forest model using the **randomForest** package. By default, the model relies on the default parameters of `randomForest::randomForest()`, except for the number of trees, which is set to 200 (`ntree = 200`). However, users can override this and train the model with custom parameter values:

```{r}
library(randomForest)
my_model <- function(X, y, ...) {
    param <- list(...)
    args <- list(x = X, y = y, ntree = 200) # default parameters
    args <- args[!(names(args) %in% names(param))]
    args <- c(args, param)
    do.call(randomForest::randomForest, args = args)
}
# The random forest is built with our default parameters
m <- create_model(USAccDeaths, method = my_model)
print(m$model$ntree)
print(m$model$mtry)
# The random forest is built with ntree = 400 and mtry = 6
m <- create_model(USAccDeaths, method = my_model, ntree = 400, mtry = 6)
print(m$model$ntree)
print(m$model$mtry)
```

# Estimating forecast accuracy

This section demonstrates how to estimate the forecast accuracy of a regression model when predicting a time series with **utsf**. Below is an example:

```{r}
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- efa(m, h = 4, type = "normal", size = 8)
r$per_horizon
r$global
```

To assess model forecast accuracy, use the `efa()` function on the fitted model. In this example, the forecasting method is a $k$-nearest neighbors algorithm (`method = "knn"`) with autoregressive lags 1 to 4, applied to the `UKgas` time series. The function estimates forecast accuracy across a 4-step-ahead horizon (`h = 4`) using multiple evaluation metrics. 

The `efa()` function returns a list with two components:

* `per_horizon`: Contains accuracy estimates for each individual forecasting step.
* `global`: Contains the overall accuracy estimates, calculated as the row means of the `per_horizon` matrix.

Currently, the following forecast accuracy measures are computed:

* MAE: mean absolute error
* MAPE: mean absolute percentage error
* sMAPE: symmetric MAPE
* RMSE: root mean squared error

Next, we describe how the forecasting accuracy measures are computed for a forecasting horizon $h$ ($y_t$ and $\hat{y}_t$ are the actual future value and its forecast for horizon $t$ respectively):

$$
MAE = \frac{1}{h}\sum_{t=1}^{h} |y_t-\hat{y}_t|
$$

$$
MAPE = \frac{1}{h}\sum_{t=1}^{h} 100\frac{|y_t-\hat{y}_t|}{y_t}
$$
$$
sMAPE = \frac{1}{h}\sum_{t=1}^{h} 200\frac{\left|y_{t}-\hat{y}_{t}\right|}{|y_t|+|\hat{y}_t|}
$$

$$
RMSE = \sqrt{\frac{1}{h}\sum_{t=1}^{h} (y_t-\hat{y}_t)^2}
$$

Forecast accuracy is estimated using the well-known rolling origin evaluation strategy. This technique generates multiple training and test sets from the time series, as illustrated below for a 4-step-ahead forecasting horizon (`h = 4`):

```{r out.width = '85%', echo = FALSE}
knitr::include_graphics("ro_normal.png")
```

In this setup, the last nine observations of the time series are reserved to construct the test sets. Six models are fitted using the expanding training sets, and their forecasts are evaluated against the corresponding test values. Consequently, the forecast error for each step of the horizon is averaged across six evaluation folds.

The number of trailing observations used to form the test sets can be controlled via the `size` or `prop` parameters:

```{r, eval = FALSE}
m <- create_model(UKgas, lags = 1:4, method = "knn")
# Use the last 9 observations of the series to build test sets
r1 <- efa(m, h = 4, type = "normal", size = 9)
# Use the last 20% observations of the series to build test sets
r2 <- efa(m, h = 4, type = "normal", prop = 0.2)
```

If the series is short, the training sets may be too small to fit a meaningful model. In such cases, an alternative rolling origin scheme can be applied by setting the `type` parameter to `"minimum"`:

```{r}
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- efa(m, h = 4, type = "minimum")
r$per_horizon
r$global
```

When this option is selected, the last `h` observations (4 in this example) are used to construct the test sets, as illustrated below for `h = 4`:

```{r out.width = '85%', echo = FALSE}
knitr::include_graphics("ro_minimum.png")
```

In this case, the estimated forecast error for horizon 1 is based on 4 errors, for horizon 2 on 3 errors, for horizon 3 on 2 errors, and for horizon 4 on 1 error.

The `efa()` function also returns the test sets (and their associated predictions) used during accuracy assessment:

```{r}
timeS <- ts(1:25)
m <- create_model(timeS, lags = 1:3, method = "mt")
r <- efa(m, h = 5, size = 7)
r$test_sets
r$predictions
```

Each row in these tables represents a test set or a set of predictions. Below are the test sets generated when using the `"minimum"` evaluation scheme:

```{r}
timeS <- ts(1:25)
m <- create_model(timeS, lags = 1:3, method = "mt")
r <- efa(m, h = 3, type = "minimum")
r$test_sets
r$predictions
```

# Parameter tuning

Another key feature of **utsf** is parameter tuning. The `tune_grid()` function estimates the forecast accuracy of a model across various parameter combinations. The optimal configuration (the one achieving the best forecast accuracy) is then used to train a final model on the full historical series and generate future forecasts. Below is an example:

```{r}
m <- create_model(UKgas, lags = 1:4, method = "knn")
r <- tune_grid(m, h = 4, tuneGrid = expand.grid(k = 1:7), type = "normal", size = 8)
# To check the estimated forecast accuracy with the different configurations
r$tuneGrid
```

In this example, the `tuneGrid` parameter specifies (as a data frame) the set of hyperparameters to evaluate. The forecast accuracy for each parameter combination is estimated as described in the previous section, using the most recent observations of the time series as a validation set. The `size` parameter controls the length of this test set. 

The `tuneGrid` component of the list returned by `tune_grid()` contains these estimation results. Here, the $k$-nearest neighbors model with $k = 1$ achieves the best performance across all forecast accuracy metrics, which is driven by the increasing variance of the series. The optimal parameter combination according to RMSE is then used to generate the final time series forecast:

```{r}
r$best
r$forecast
```

Below, we plot the $k$ values against their estimated forecast accuracy in terms of RMSE:

```{r}
plot(r$tuneGrid$k, r$tuneGrid$RMSE, type = "o", pch = 19, xlab = "k (number of nearest neighbors)", ylab = "RMSE", main = "Estimated accuracy")
```

The following example illustrates tuning a model with multiple hyperparameters. Here, we evaluate a random forest model using three tuning parameters:

```{r}
m <- create_model(UKDriverDeaths, lags = 1:12, method = "rf")
r <- tune_grid(m, 
               h = 12, 
               tuneGrid = expand.grid(num.trees = c(200, 500), replace = c(TRUE, FALSE), mtry = c(4, 8)),
               type = "normal", 
               size = 12
)
r$tuneGrid
```

# Preprocessings and transformations

Model forecast accuracy can often be improved by applying transformations or preprocessing techniques to the target time series. Currently, **utsf** focuses on preprocessing methods designed for trended time series. This is because most models currently integrated into the package (such as regression trees and $k$-nearest neighbors) predict averages of the target values in the training set (i.e., historical observations). When a series exhibits a trend, these historical averages typically fall outside the range of future values. 

The following example demonstrates forecasting a trended series (the annual revenue passenger miles flown by commercial airlines in the United States) using a random forest model:

```{r}
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "none")
f <- forecast(m, h = 4)
autoplot(f)
```

In this case, no preprocessing is applied to handle the trend (`trend = "none"`). As shown below, the forecast fails to capture the upward trajectory of the series because regression tree models can only predict average values of the training targets.

The following subsections explain how to handle trended series using three different transformations:

## Differencing

A common approach for handling trended series is first-differencing (computing the differences between consecutive observations). The model is then trained on the differenced series, and the resulting forecasts are back-transformed to the original scale. Below is an example:

```{r}
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "differences", nfd = 1)
f <- forecast(m, h = 4)
autoplot(f)
```

In this example, the `trend` parameter is set to `"differences"` to apply first-differencing. The order of differencing is specified via `nfd` (which typically defaults to 1). Setting `nfd = -1` automatically estimates the required order of differencing using the `ndiffs()` function from the **forecast** package. In this case, if the estimated order is 0, no differencing is applied.

```{r}
# The order of first differences is estimated using the ndiffs function
m <- create_model(airmiles, lags = 1:4, method = "rf", trend = "differences", nfd = -1)
m$differences
```

## The additive transformation

This transformation is also used for dealing with trended series in other R packages, such as **tsfknn** or **tsfgrnn**. The additive transformation modifies the targets of the training set as follows:

* The target associated with a feature vector is transformed by subtracting the mean of that feature vector.
* To back-transform a prediction, the mean of the corresponding input vector is added to it.

The effect of the additive transformation on the training examples can be easily inspected using the package API. For instance, consider the training set of a model fit without preprocessing:

```{r}
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "none")
cbind(m$features, Targets = m$targets)
```

Now, consider the effect of the additive transformation on the training examples:

```{r}
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "additive", transform_features = FALSE)
cbind(m$features, Targets = m$targets)
```

In addition to transforming the targets, the features can also be transformed. A feature vector is transformed by subtracting its mean from each element. The effect of this transformation is shown below:

```{r}
timeS <- ts(c(1, 3, 7, 9, 10, 12))
m <- create_model(timeS, lags = 1:2, trend = "additive", transform_features = TRUE)
cbind(m$features, Targets = m$targets)
```

Transforming the features removes level effects from the feature space, which empirically improves forecast accuracy. 

Below, we forecast the `airmiles` time series using the additive transformation with a random forest model:

```{r}
m <- create_model(airmiles, lags = 1:4, method = "rf")
f <- forecast(m, h = 4)
autoplot(f)
```

By default, the additive transformation is applied to both targets and features.

## The multiplicative transformation

The multiplicative transformation is similar to the additive transformation, but it is designed for series with exponential trends, whereas the additive transformation is better suited for linear trends. Under the multiplicative transformation, training examples are processed as follows:

* The target associated with a feature vector is transformed by dividing it by the mean of that feature vector.
* To back-transform a prediction, it is multiplied by the mean of the corresponding input vector.
* Optionally, each element of a feature vector is transformed by dividing it by the vector's mean.

Below is an example using a synthetic time series with an exponential trend, comparing forecasts generated with the additive and multiplicative transformations:

```{r}
t <- ts(10 * 1.05^(1:20))
m_m <- create_model(t, lags = 1:3, method = "rf", trend = "multiplicative")
f_m <- forecast(m_m, h = 4)
m_a <- create_model(t, lags = 1:3, method = "rf", trend = "additive")
f_a <- forecast(m_a, h = 4)
library(vctsfr)
plot_predictions(t, predictions = list(Multiplicative = f_m$pred, Additive = f_a$pred))
```

In this case, the forecast generated with the multiplicative transformation captures the exponential trend perfectly.

As an additional example, we compare the additive and multiplicative transformations applied to the `AirPassengers` series (a monthly series exhibiting an exponential trend):

```{r, fig.width= 6}
m_m <- create_model(AirPassengers, lags = 1:12, method = "knn", trend = "multiplicative")
f_m <- forecast(m_m, h = 36)
m_a <- create_model(AirPassengers, lags = 1:12, method = "knn", trend = "additive")
f_a <- forecast(m_a, h = 36)
library(vctsfr)
plot_predictions(AirPassengers, predictions = list(Multiplicative = f_m$pred, Additive = f_a$pred))
```

# Default parameters

The `create_model()` function requires only one mandatory parameter: the time series used to train the model. Below, we describe the default settings for the remaining parameters:

* `lags`: An integer vector defining the autoregressive lags. If `frequency(ts) == f` (where `ts` is the input time series and $f > 1$), the default lags are `1:f` (e.g., `1:4` for quarterly data and `1:12` for monthly data), capturing potential seasonal patterns. If `frequency(ts) == 1`:
    * Significant lags are automatically selected based on the partial autocorrelation function (`stats::pacf()`).
    * If no lag is statistically significant, lags `1:5` are selected.
    * If only a single lag is significant and the additive or multiplicative transformation is applied to both targets and features, lags `1:5` are used instead. This prevents applying feature-level transformations when only one lag is present, as scaling a single feature by its own mean is uninformative.
* `method`: Defaults to the $k$-nearest neighbors algorithm.
* `trend`: Defaults to the additive transformation (`"additive"`). This method has proven effective for trended time series and generally improves performance across most series types.
* `transform_features`: Defaults to `TRUE`, ensuring that both features and targets are transformed when using the additive or multiplicative transformation.

# More on using your custom models

In previous sections we have explained how to leverage the package's features to 
apply your own models for time series forecasting. In this section we describe
how to take advantage of other package's features. We will employ the random forest
model implemented in the **randomForest** package used previously. First, let us
examine the computation of prediction intervals:

```{r}
library(randomForest)
myModel <- function(X, y, ...) {
  param <- list(...)
  args <- list(x = X, y = y, ntree = 200)
  args <- args[!(names(args) %in% names(param))]
  args <- c(args, param)
  do.call(randomForest::randomForest, args = args)
}
m <- create_model(USAccDeaths, method = myModel)
f <- forecast(m, h = 12, level = 90, PI = TRUE)
f
```

As can be seen, all these features are used in the same way as with the built-in models. For example, let us estimate the forecast accuracy of the previous model for a forecasting horizon of 5:

```{r}
efa(m, h = 5, type = "normal", size = 8)$per_horizon
```

By default, the additive transformation is applied. Let us select the multiplicative transformation to estimate the forecast accuracy under this setting:

```{r}
m <- create_model(USAccDeaths, lags = 1:12, method = myModel,
	                    trend = "multiplicative")
efa(m, h = 5, type = "normal", size = 8)$per_horizon
```

Finally, we perform parameter tuning over different values for the `mtry` and 
`ntree` arguments:

```{r}
m <- create_model(USAccDeaths, method = myModel)
r <- tune_grid(m, h = 6,
               tuneGrid = expand.grid(mtry = 3:6, ntree = c(200, 500)),
               type = "normal"
)
r$tuneGrid
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate_accuracy.R
\name{efa}
\alias{efa}
\title{Estimate the forecast accuracy of a model on a time series}
\usage{
efa(model, h, type = c("normal", "minimum"), size = NULL, prop = NULL)
}
\arguments{
\item{model}{An object of class \code{utsf} with a model trained with a time
series.}

\item{h}{A positive integer. The forecasting horizon.}

\item{type}{A string. Possible values are \code{"normal"} (the default) and
\code{"minimum"}. See the vignette \link{utsf} for an explanation of both ways
of evaluating forecast accuracy.}

\item{size}{An integer. It is the size of the test set (how many of the last
observations of the time series are used as test set). It can only be used
when the type parameter is \code{"normal"}. By default, it is the length of the
forecasting horizon.}

\item{prop}{A numeric value in the range (0, 1). It is the proportion of the
time series used as test set. It can only be used when the type parameter is
\code{"normal"}.}
}
\value{
A list with four components: \item{\code{per_horizon}}{A matrix with the
estimated forecast accuracy per forecasting horizon using several
forecasting accuracy measures.}
\item{\code{global}}{The average estimated forecast accuracy for all the horizons. It is computed as the mean
of the different rows of the \code{per_horizon} component.}
\item{\code{test_sets}}{A matrix with the test sets used in the evaluation. Each
row of the matrix is a test set.} \item{\code{predictions}}{The predictions for the test sets.}
}
\description{
It uses an object of class \code{utsf} to asses the forecasting accuracy of its
associated model on its associated time series applying a rolling origin
evaluation.
}
\examples{
m <- create_model(UKgas, lags = 1:4, method = "rt")
efa(m, h = 4, type = "normal", size = 8)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utsf-package.R
\docType{package}
\name{utsf-package}
\alias{utsf}
\alias{utsf-package}
\title{utsf: Univariate Time Series Forecasting}
\description{
An engine for univariate time series forecasting using different regression models in an autoregressive way. The engine provides an uniform interface for applying the different models. Furthermore, it is extensible so that users can easily apply their own regression models to univariate time series forecasting and benefit from all the features of the engine, such as preprocessings or estimation of forecast accuracy.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/franciscomartinezdelrio/utsf}
  \item Report bugs at \url{https://github.com/franciscomartinezdelrio/utsf/issues}
}

}
\author{
\strong{Maintainer}: Francisco Martinez \email{fmartin@ujaen.es} (\href{https://orcid.org/0000-0002-5206-1898}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Maria Pilar Frias-Bustamante \email{mpfrias@ujaen.es} (\href{https://orcid.org/0000-0001-6886-0953}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_examples.R
\name{build_examples}
\alias{build_examples}
\title{Build the training examples}
\usage{
build_examples(timeS, lags)
}
\arguments{
\item{timeS}{The time series.}

\item{lags}{An integer vector with the lags used as feature vector in
decreasing order.}
}
\value{
A list with two fields: 1) a matrix with the features of the examples
and 2) a vector with the targets of the examples
}
\description{
Build the training examples for a regressive model to forecast a time series
using lagged values of the series as autoregressive features.
}
\examples{
build_examples(ts(1:5), lags = 2:1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.utsf_forecast}
\alias{autoplot.utsf_forecast}
\title{Create a ggplot object from an \code{utsf_forecast} object}
\usage{
\method{autoplot}{utsf_forecast}(object, ...)
}
\arguments{
\item{object}{An object of class \code{utsf_forecast}.}

\item{...}{additional parameter.}
}
\value{
The \code{ggplot} object representing a plotting of the time series and
its forecast.
}
\description{
Plot the time series and its associated forecast.
}
\examples{
m <- create_model(AirPassengers, lags = 1:12, method = "rf")
f <- forecast(m, h = 12)
library(ggplot2)
autoplot(f)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{create_model}
\alias{create_model}
\title{Train an univariate time series forecasting model}
\usage{
create_model(
  timeS,
  lags = NULL,
  method = c("knn", "lm", "rt", "mt", "bagging", "rf", "xgboost"),
  trend = c("additive", "multiplicative", "differences", "none"),
  nfd = -1,
  transform_features = TRUE,
  ...
)
}
\arguments{
\item{timeS}{A time series of class \code{ts} or a numeric vector.}

\item{lags}{An integer vector, in increasing order, expressing the lags used
as autoregressive variables. If the default value (\code{NULL}) is provided, a
suitable vector is chosen.}

\item{method}{A string indicating the method used for training and
forecasting. Allowed values are:
\itemize{
\item \code{"knn"}: k-nearest neighbors (the default)
\item \code{"lm"}: linear regression
\item \code{"rt"}: regression trees
\item \code{"mt"}:  model trees
\item \code{"bagging"}
\item \code{"rf"}: random forests
\item \code{"xgboost"}: extreme gradient boosting
}

See details for a brief explanation of the models. It is also possible to
use your own regression model, in that case a function explaining how to
build your model must be provided, see the vignette for further details.}

\item{trend}{A character indicating the type of preprocessing applied to the
time series in order to deal with trending series, see the vignette for
details.}

\item{nfd}{In the case that the parameter \code{trend} has the value "differences",
it specifies the order of first differences to be applied. If the default
(-1) is used, the order of first differences needed by the time series will
be estimated by the \code{\link[forecast:ndiffs]{forecast::ndiffs()}} function.}

\item{transform_features}{A logical value indicating whether the training
features are also transformed if the additive or multiplicative
transformation has been used as preprocessing to deal with trending series.}

\item{...}{Parameters for the underlying function that builds the model. If no
parameters are provided, the model is normally fitted with its default
parameters. See details for the functions used to train the models.}
}
\value{
An S3 object of class \code{utsf}, basically a list with, at least, the
following components: \item{\code{ts}}{The time series being forecast.}
\item{\code{features}}{A data frame with the features of the training set. The
column names of the data frame indicate the autoregressive lags.}
\item{\code{targets}}{A vector with the targets of the training set.}
\item{\code{lags}}{An integer vector with the autoregressive lags.}
\item{\code{model}}{The regression model used recursively to make the forecast.}
}
\description{
This function trains a model from the historical values of a time series using
an autoregressive approach: the targets are the historical values and the
features of the targets their lagged values.
}
\details{
The functions used to build and train the model are:
\itemize{
\item KNN: In this case no model is built and the function \code{\link[FNN:knn.reg]{FNN::knn.reg()}} is
used to predict the future values of the time series. By default, k is equal
to 3.
\item Linear models: Function \code{\link[stats:lm]{stats::lm()}} to build the model and the method
\code{\link[stats:predict.lm]{stats::predict.lm()}} associated with the trained model to forecast the future
values of the time series.
\item Regression trees: Function \code{\link[rpart:rpart]{rpart::rpart()}} to build the model and the
method \code{\link[rpart:predict.rpart]{rpart::predict.rpart()}} associated with the trained model to forecast
the future values of the time series.
\item Model trees: Function \code{\link[Cubist:cubist.default]{Cubist::cubist()}} to build the model and the
method \code{\link[Cubist:predict.cubist]{Cubist::predict.cubist()}} associated with the trained model to
forecast the future values of the time series. By default, the parameter
\code{committees} is set to 5.
\item Bagging: Function \code{\link[ipred:bagging]{ipred::bagging()}} to build the model and the
method \code{\link[ipred:predict.bagging]{ipred::predict.regbagg()}} associated with the trained model to
forecast the future values of the time series.
\item Random forest: Function \code{\link[ranger:ranger]{ranger::ranger()}} to build the model and the
method \code{\link[ranger:predict.ranger]{ranger::predict.ranger()}} associated with the trained model to
forecast the future values of the time series.
\item Extreme gradient boosting: Function \code{\link[xgboost:xgboost]{xgboost::xgboost()}} to build the model and the
method \code{\link[xgboost:predict.xgboost]{xgboost::predict.xgboost()}} associated with the trained model to
forecast the future values of the time series.
}
}
\examples{
## Build model using k-nearest neighbors
create_model(AirPassengers, method = "knn")

## Using k-nearest neighbors changing the default k value
create_model(AirPassengers, method = "knn", k = 5)

## Using your own regression model

# Function to build the regression model
my_knn_model <- function(X, y, param) {
  structure(list(X = X, y = y), class = "my_knn")
}
# Function to predict a new example
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
}
create_model(AirPassengers, method = my_knn_model)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/main.R
\name{forecast.utsf}
\alias{forecast.utsf}
\title{Forecasting a time series}
\usage{
\method{forecast}{utsf}(object, h, PI = FALSE, level = 90, ...)
}
\arguments{
\item{object}{an object of class \code{utsf} embedding a forecasting model for a
time series.}

\item{h}{A positive integer. Number of values to be forecast into the future,
i.e., forecast horizon.}

\item{PI}{If TRUE, prediction intervals are produced using simulation and
assuming normally distributed errors.}

\item{level}{Confidence level for predictions intervals.}

\item{...}{Other arguments passed to methods}
}
\value{
an object of class \code{utsf_forecast} with the same components of the
model received as first argument, plus several components:
\item{\code{pred}}{The forecast as an \code{ts} object.}
\item{\code{lower}}{Lower limits for prediction interval.}
\item{\code{upper}}{Upper limits for prediction interval.}
\item{\code{level}}{Confidence value associated with the prediction interval}
}
\description{
Forecasting a time series
}
\examples{
## Forecast time series using k-nearest neighbors
m <- create_model(USAccDeaths, method = "knn")
f <- forecast(m, h = 12)
f$pred
library(ggplot2)
autoplot(f)

## Using k-nearest neighbors changing the default k value
m <- create_model(USAccDeaths, method = "knn", k = 5)
forecast(m, h = 12)

## Using your own regression model

# Function to build the regression model
my_knn_model <- function(X, y, param) {
  structure(list(X = X, y = y), class = "my_knn")
}
# Function to predict a new example
predict.my_knn <- function(object, new_value) {
  FNN::knn.reg(train = object$X, test = new_value, y = object$y)$pred
}
m <- create_model(USAccDeaths, method = my_knn_model)
forecast(m, h = 12)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/do_tuneGrid.R
\name{tune_grid}
\alias{tune_grid}
\title{Estimate the forecast accuracy of a model on a time series according to a grid
of parameters}
\usage{
tune_grid(
  model,
  h,
  tuneGrid,
  type = c("normal", "minimum"),
  size = NULL,
  prop = NULL
)
}
\arguments{
\item{model}{An object of class \code{utsf} with a model trained with a time
series.}

\item{h}{A positive integer. The forecasting horizon.}

\item{tuneGrid}{A data frame with possible tuning values. The columns are
named as the tuning parameters.}

\item{type}{A string. Possible values are \code{"normal"} (the default) and
\code{"minimum"}. See the vignette \link{utsf} for an explanation of both ways
of evaluating forecast accuracy.}

\item{size}{An integer. It is the size of the test set (how many of the last
observations of the time series are used as test set). It can only be used
when the type parameter is \code{"normal"}. By default, it is the length of the
forecasting horizon.}

\item{prop}{A numeric value in the range (0, 1). It is the proportion of the
time series used as test set. It can only be used when the type parameter is
\code{"normal"}.}
}
\value{
A list with three components: \item{\code{tuneGrid}}{A data frame with the
different combination of parameters and the estimated forecast accuracy of a
model trained with those parameters.} \item{\code{best}}{The best combination of
parameters according to root mean squared error.} \item{\code{forecast}}{An
object of class \code{utsf_forecast} with the forecast for horizon \code{h} using the
best estimated combination of parameters.}
}
\description{
It uses an object of class \code{utsf} to asses the forecasting accuracy of its
associated model on its associated time series applying rolling origin
evaluation according to different configurations of model parameters.
}
\details{
The estimation of forecast accuracy is done with the \code{\link[=efa]{efa()}} function. The
best combination of parameters is used to train the model with all the
historical values of the time series and forecast \code{h} values ahead.
}
\examples{
m <- create_model(UKgas, lags = 1:4, method = "knn")
tune_grid(m, h = 4, tuneGrid = expand.grid(k = 1:7), type = "normal", size = 8)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/build_models.R
\name{predict.utsf}
\alias{predict.utsf}
\title{Prediction from \code{utsf} objects}
\usage{
\method{predict}{utsf}(object, new_value, ...)
}
\arguments{
\item{object}{object of class \code{utsf}.}

\item{new_value}{a data frame with one row of a new observation.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
a numeric value with the forecast.
}
\description{
Predict the class of a new observation based on the model associated with the
\code{utsf} object
}
